"""
Deployment coordinator for managing deployment lifecycle.
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from contextlib import AsyncExitStack
from dataclasses import dataclass
from datetime import UTC, datetime
from uuid import UUID

from ai.backend.common.clients.http_client.client_pool import ClientPool
from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.data.endpoint.types import EndpointLifecycle
from ai.backend.common.data.notification import NotificationRuleType
from ai.backend.common.data.notification.messages import EndpointLifecycleChangedMessage
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.notification import NotificationTriggeredEvent
from ai.backend.common.events.event_types.schedule.anycast import (
    DoDeploymentLifecycleEvent,
    DoDeploymentLifecycleIfNeededEvent,
)
from ai.backend.common.leader.tasks.event_task import EventTaskSpec
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.deployment.types import DeploymentInfo
from ai.backend.manager.data.session.types import SchedulingResult
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.repositories.base.creator import BulkCreator
from ai.backend.manager.repositories.base.updater import BatchUpdater
from ai.backend.manager.repositories.deployment import (
    DeploymentConditions,
    DeploymentRepository,
)
from ai.backend.manager.repositories.deployment.creators import EndpointLifecycleBatchUpdaterSpec
from ai.backend.manager.repositories.scheduling_history.creators import DeploymentHistoryCreatorSpec
from ai.backend.manager.sokovan.deployment.recorder import DeploymentRecorderContext
from ai.backend.manager.sokovan.deployment.route.route_controller import RouteController
from ai.backend.manager.sokovan.recorder.types import ExecutionRecord
from ai.backend.manager.sokovan.recorder.utils import extract_sub_steps_for_entity
from ai.backend.manager.sokovan.scheduling_controller.scheduling_controller import (
    SchedulingController,
)
from ai.backend.manager.types import DistributedLockFactory

from .deployment_controller import DeploymentController
from .executor import DeploymentExecutor
from .handlers import (
    CheckPendingDeploymentHandler,
    CheckReplicaDeploymentHandler,
    DeploymentHandler,
    DestroyingDeploymentHandler,
    ReconcileDeploymentHandler,
    ScalingDeploymentHandler,
)
from .types import DeploymentExecutionResult, DeploymentLifecycleType

log = BraceStyleAdapter(logging.getLogger(__name__))


@dataclass
class DeploymentTaskSpec:
    """Specification for a deployment lifecycle periodic task."""

    lifecycle_type: DeploymentLifecycleType
    short_interval: float | None = None  # None means no short-cycle task
    long_interval: float = 60.0
    initial_delay: float = 30.0

    def create_if_needed_event(self) -> DoDeploymentLifecycleIfNeededEvent:
        """Create event for checking if processing is needed."""
        return DoDeploymentLifecycleIfNeededEvent(self.lifecycle_type.value)

    def create_process_event(self) -> DoDeploymentLifecycleEvent:
        """Create event for forced processing."""
        return DoDeploymentLifecycleEvent(self.lifecycle_type.value)

    @property
    def short_task_name(self) -> str:
        """Name for the short-cycle task."""
        return f"deployment_process_if_needed_{self.lifecycle_type.value}"

    @property
    def long_task_name(self) -> str:
        """Name for the long-cycle task."""
        return f"deployment_process_{self.lifecycle_type.value}"


class DeploymentCoordinator:
    """Coordinates deployment-related operations."""

    _valkey_schedule: ValkeyScheduleClient
    _deployment_controller: DeploymentController
    _deployment_repository: DeploymentRepository
    _deployment_handlers: Mapping[DeploymentLifecycleType, DeploymentHandler]
    _lock_factory: DistributedLockFactory
    _config_provider: ManagerConfigProvider
    _event_producer: EventProducer

    def __init__(
        self,
        valkey_schedule: ValkeyScheduleClient,
        deployment_controller: DeploymentController,
        deployment_repository: DeploymentRepository,
        event_producer: EventProducer,
        lock_factory: DistributedLockFactory,
        config_provider: ManagerConfigProvider,
        scheduling_controller: SchedulingController,
        client_pool: ClientPool,
        valkey_stat: ValkeyStatClient,
        route_controller: RouteController,
    ) -> None:
        """Initialize the deployment coordinator."""
        self._valkey_schedule = valkey_schedule
        self._deployment_controller = deployment_controller
        self._deployment_repository = deployment_repository
        self._event_producer = event_producer
        self._lock_factory = lock_factory
        self._config_provider = config_provider
        self._route_controller = route_controller

        # Create deployment executor
        executor = DeploymentExecutor(
            deployment_repo=self._deployment_repository,
            scheduling_controller=scheduling_controller,
            config_provider=self._config_provider,
            client_pool=client_pool,
            valkey_stat=valkey_stat,
        )
        self._deployment_handlers = self._init_handlers(executor)

    def _init_handlers(
        self, executor: DeploymentExecutor
    ) -> Mapping[DeploymentLifecycleType, DeploymentHandler]:
        """Initialize and return the mapping of deployment lifecycle types to their handlers."""
        return {
            DeploymentLifecycleType.CHECK_PENDING: CheckPendingDeploymentHandler(
                deployment_executor=executor,
                deployment_controller=self._deployment_controller,
            ),
            DeploymentLifecycleType.CHECK_REPLICA: CheckReplicaDeploymentHandler(
                deployment_executor=executor,
                deployment_controller=self._deployment_controller,
            ),
            DeploymentLifecycleType.SCALING: ScalingDeploymentHandler(
                deployment_executor=executor,
                deployment_controller=self._deployment_controller,
                route_controller=self._route_controller,
            ),
            DeploymentLifecycleType.RECONCILE: ReconcileDeploymentHandler(
                deployment_executor=executor,
                deployment_controller=self._deployment_controller,
            ),
            DeploymentLifecycleType.DESTROYING: DestroyingDeploymentHandler(
                deployment_executor=executor,
                deployment_controller=self._deployment_controller,
                route_controller=self._route_controller,
            ),
        }

    async def process_deployment_lifecycle(
        self,
        lifecycle_type: DeploymentLifecycleType,
    ) -> None:
        handler = self._deployment_handlers.get(lifecycle_type)
        if not handler:
            log.warning("No handler for deployment lifecycle type: {}", lifecycle_type.value)
            return
        async with AsyncExitStack() as stack:
            if handler.lock_id is not None:
                lock_lifetime = self._config_provider.config.manager.session_schedule_lock_lifetime
                await stack.enter_async_context(self._lock_factory(handler.lock_id, lock_lifetime))
            deployments = await self._deployment_repository.get_endpoints_by_statuses(
                handler.target_statuses()
            )
            if not deployments:
                log.trace("No deployments to process for handler: {}", handler.name())
                return
            log.info("handler: {} - processing {} deployments", handler.name(), len(deployments))

            # Execute handler with recorder context
            deployment_ids = [d.id for d in deployments]
            with DeploymentRecorderContext.scope(
                lifecycle_type.value, entity_ids=deployment_ids
            ) as pool:
                result = await handler.execute(deployments)
                all_records = pool.build_all_records()

                # Handle status transitions with history recording
                await self._handle_status_transitions(handler, result, all_records)

            try:
                await handler.post_process(result)
            except Exception as e:
                log.error("Error during post-processing: {}", e)

    async def _handle_status_transitions(
        self,
        handler: DeploymentHandler,
        result: DeploymentExecutionResult,
        records: Mapping[UUID, ExecutionRecord],
    ) -> None:
        """Handle status transitions with history recording.

        All transitions (success and failure) are processed in a single transaction
        to ensure atomicity.

        Args:
            handler: The deployment handler that was executed
            result: The result of the handler execution
            records: Execution records from the recorder context
        """
        handler_name = handler.name()
        target_statuses = handler.target_statuses()
        from_status = target_statuses[0] if target_statuses else None

        # Collect all batch updaters and history specs
        batch_updaters: list[BatchUpdater[EndpointRow]] = []
        all_history_specs: list[DeploymentHistoryCreatorSpec] = []
        notification_events: list[NotificationTriggeredEvent] = []
        timestamp_now = datetime.now(UTC).isoformat()

        # Handle success transitions
        next_status = handler.next_status()
        if next_status is not None and result.successes:
            endpoint_ids = [d.id for d in result.successes]
            success_history_specs = [
                DeploymentHistoryCreatorSpec(
                    deployment_id=d.id,
                    phase=handler_name,
                    result=SchedulingResult.SUCCESS,
                    message=f"{handler_name} completed successfully",
                    from_status=from_status,
                    to_status=next_status,
                    sub_steps=extract_sub_steps_for_entity(d.id, records),
                )
                for d in result.successes
            ]
            batch_updaters.append(
                BatchUpdater(
                    spec=EndpointLifecycleBatchUpdaterSpec(lifecycle_stage=next_status),
                    conditions=[
                        DeploymentConditions.by_ids(endpoint_ids),
                        DeploymentConditions.by_lifecycle_stages(target_statuses),
                    ],
                )
            )
            all_history_specs.extend(success_history_specs)
            notification_events.extend([
                self._build_lifecycle_notification_event(
                    deployment=d,
                    from_status=from_status,
                    to_status=next_status,
                    transition_result="success",
                    timestamp=timestamp_now,
                )
                for d in result.successes
            ])

        # Handle failure transitions
        failure_status = handler.failure_status()
        if failure_status is not None and result.errors:
            endpoint_ids = [e.deployment_info.id for e in result.errors]
            failure_history_specs = [
                DeploymentHistoryCreatorSpec(
                    deployment_id=e.deployment_info.id,
                    phase=handler_name,
                    result=SchedulingResult.FAILURE,
                    message=e.reason,
                    from_status=from_status,
                    to_status=failure_status,
                    error_code=e.error_code,
                    sub_steps=extract_sub_steps_for_entity(e.deployment_info.id, records),
                )
                for e in result.errors
            ]
            batch_updaters.append(
                BatchUpdater(
                    spec=EndpointLifecycleBatchUpdaterSpec(lifecycle_stage=failure_status),
                    conditions=[
                        DeploymentConditions.by_ids(endpoint_ids),
                        DeploymentConditions.by_lifecycle_stages(target_statuses),
                    ],
                )
            )
            all_history_specs.extend(failure_history_specs)
            notification_events.extend([
                self._build_lifecycle_notification_event(
                    deployment=e.deployment_info,
                    from_status=from_status,
                    to_status=failure_status,
                    transition_result="failure",
                    timestamp=timestamp_now,
                )
                for e in result.errors
            ])

        # Execute all updates in a single transaction
        if batch_updaters:
            await self._deployment_repository.update_endpoint_lifecycle_bulk_with_history(
                batch_updaters, BulkCreator(specs=all_history_specs)
            )

        # Anycast notification events
        for event in notification_events:
            try:
                await self._event_producer.anycast_event(event)
            except Exception as e:
                log.warning("Failed to send lifecycle notification: {}", e)

    def _build_lifecycle_notification_event(
        self,
        deployment: DeploymentInfo,
        from_status: EndpointLifecycle | None,
        to_status: EndpointLifecycle,
        transition_result: str,
        timestamp: str,
    ) -> NotificationTriggeredEvent:
        """Build a notification event for a lifecycle transition."""
        message = EndpointLifecycleChangedMessage(
            endpoint_id=str(deployment.id),
            endpoint_name=deployment.metadata.name,
            domain=deployment.metadata.domain,
            project_id=str(deployment.metadata.project),
            resource_group=deployment.metadata.resource_group,
            from_status=from_status.value if from_status else None,
            to_status=to_status.value,
            transition_result=transition_result,
            event_timestamp=timestamp,
        )
        return NotificationTriggeredEvent(
            rule_type=NotificationRuleType.ENDPOINT_LIFECYCLE_CHANGED.value,
            timestamp=datetime.now(UTC),
            notification_data=message.model_dump(),
        )

    async def process_if_needed(self, lifecycle_type: DeploymentLifecycleType) -> None:
        """
        Process deployment lifecycle operation if needed (based on internal state).

        Args:
            lifecycle_type: Type of deployment lifecycle operation

        Returns:
            True if operation was performed, False otherwise
        """
        # Check internal state (uses Redis marks)
        if not await self._valkey_schedule.load_and_delete_deployment_mark(lifecycle_type.value):
            return
        await self.process_deployment_lifecycle(lifecycle_type)

    @staticmethod
    def _create_task_specs() -> list[DeploymentTaskSpec]:
        """Create task specifications for all deployment lifecycle types."""
        return [
            # Check pending deployments frequently with both short and long cycles
            DeploymentTaskSpec(
                DeploymentLifecycleType.CHECK_PENDING,
                short_interval=2.0,
                long_interval=30.0,
                initial_delay=10.0,
            ),
            # Check replicas moderately with both short and long cycles
            DeploymentTaskSpec(
                DeploymentLifecycleType.CHECK_REPLICA,
                short_interval=5.0,
                long_interval=30.0,
                initial_delay=10.0,
            ),
            # Scaling operations with both short and long cycles
            DeploymentTaskSpec(
                DeploymentLifecycleType.SCALING,
                short_interval=5.0,
                long_interval=30.0,
                initial_delay=10.0,
            ),
            DeploymentTaskSpec(
                DeploymentLifecycleType.RECONCILE,
                short_interval=None,
                long_interval=30.0,
                initial_delay=10.0,
            ),
            # Check destroying deployments - only long cycle
            DeploymentTaskSpec(
                DeploymentLifecycleType.DESTROYING,
                short_interval=5.0,
                long_interval=60.0,
                initial_delay=25.0,
            ),
        ]

    def create_task_specs(self) -> list[EventTaskSpec]:
        """Create task specifications for deployment lifecycle events."""
        task_specs = self._create_task_specs()
        specs: list[EventTaskSpec] = []

        for spec in task_specs:
            # Create short-cycle task spec if specified
            if spec.short_interval is not None:
                short_spec = EventTaskSpec(
                    name=spec.short_task_name,
                    event_factory=spec.create_if_needed_event,
                    interval=spec.short_interval,
                    initial_delay=0.0,  # Start immediately for short tasks
                )
                specs.append(short_spec)

            # Create long-cycle task spec (always present)
            long_spec = EventTaskSpec(
                name=spec.long_task_name,
                event_factory=spec.create_process_event,
                interval=spec.long_interval,
                initial_delay=spec.initial_delay,
            )
            specs.append(long_spec)

        return specs



"""Base action class for revision operations."""

from typing import override

from ai.backend.common.data.permission.types import EntityType
from ai.backend.manager.actions.action import BaseAction


class RevisionOperationBaseAction(BaseAction):
    """Base action for revision promotion and rollback operations."""

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.DEPLOYMENT_REVISION



"""Project V2 GraphQL query resolvers."""

from __future__ import annotations

from uuid import UUID

import strawberry
from strawberry import Info

from ai.backend.manager.api.gql.domain_v2.types import DomainV2GQL
from ai.backend.manager.api.gql.project_v2.fetcher import (
    fetch_admin_projects,
    fetch_domain_projects,
    fetch_project,
    fetch_project_domain,
)
from ai.backend.manager.api.gql.project_v2.types import (
    DomainProjectScope,
    ProjectV2Connection,
    ProjectV2Filter,
    ProjectV2GQL,
    ProjectV2OrderBy,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext


@strawberry.field(
    description=(
        "Added in 26.2.0. Get a single project by ID. Returns an error if project is not found."
    )
)  # type: ignore[misc]
async def project_v2(
    info: Info[StrawberryGQLContext],
    project_id: UUID,
) -> ProjectV2GQL | None:
    """Get a single project by UUID.

    Args:
        info: Strawberry GraphQL context.
        project_id: UUID of the project to retrieve.

    Returns:
        ProjectV2GQL object.

    Raises:
        ProjectNotFound: If the project with the given UUID does not exist.
    """
    return await fetch_project(info, project_id=project_id)


@strawberry.field(
    description=(
        "Added in 26.2.0. List all projects with filtering and pagination (admin only). "
        "Requires superadmin privileges."
    )
)  # type: ignore[misc]
async def admin_projects_v2(
    info: Info[StrawberryGQLContext],
    filter: ProjectV2Filter | None = None,
    order_by: list[ProjectV2OrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ProjectV2Connection | None:
    """List all projects with optional filtering, ordering, and pagination.

    Args:
        info: Strawberry GraphQL context.
        filter: Optional filter criteria.
        order_by: Optional ordering specification.
        before: Cursor for backward pagination.
        after: Cursor for forward pagination.
        first: Number of items from the start.
        last: Number of items from the end.
        limit: Maximum number of items (offset-based).
        offset: Starting position (offset-based).

    Returns:
        ProjectV2Connection with paginated project records.
    """
    return await fetch_admin_projects(
        info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(
    description=(
        "Added in 26.2.0. List projects within a specific domain. "
        "Requires domain admin privileges or higher."
    )
)  # type: ignore[misc]
async def domain_projects_v2(
    info: Info[StrawberryGQLContext],
    scope: DomainProjectScope,
    filter: ProjectV2Filter | None = None,
    order_by: list[ProjectV2OrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> ProjectV2Connection | None:
    """List projects within a specific domain.

    Args:
        info: Strawberry GraphQL context.
        scope: Domain scope specifying which domain to query.
        filter: Optional additional filter criteria.
        order_by: Optional ordering specification.
        before: Cursor for backward pagination.
        after: Cursor for forward pagination.
        first: Number of items from the start.
        last: Number of items from the end.
        limit: Maximum number of items (offset-based).
        offset: Starting position (offset-based).

    Returns:
        ProjectV2Connection with paginated project records from the domain.
    """
    from ai.backend.manager.repositories.group.types import DomainProjectSearchScope

    return await fetch_domain_projects(
        info,
        scope=DomainProjectSearchScope(domain_name=scope.domain_name),
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(
    description=(
        "Added in 26.2.0. Get the domain that a project belongs to. "
        "Returns an error if project or domain is not found."
    )
)  # type: ignore[misc]
async def project_domain_v2(
    info: Info[StrawberryGQLContext],
    project_id: UUID,
) -> DomainV2GQL | None:
    """Get the domain that a project belongs to.

    Args:
        info: Strawberry GraphQL context.
        project_id: UUID of the project.

    Returns:
        DomainV2GQL object for the project's domain.

    Raises:
        ProjectNotFound: If the project with the given UUID does not exist.
        DomainNotFound: If the project's domain does not exist.
    """
    return await fetch_project_domain(info, project_id=project_id)



"""Action to list all available export reports."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.repositories.base.export import ReportDef

from .base import ExportAction


@dataclass
class ListReportsAction(ExportAction):
    """Action to list all available export reports."""

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class ListReportsActionResult(BaseActionResult):
    """Result of listing export reports."""

    reports: list[ReportDef]

    @override
    def entity_id(self) -> str | None:
        return None



"""
PBKDF2-based password hashing implementation.
"""

from __future__ import annotations

import hashlib
import secrets

from ai.backend.manager.data.auth.hash import PasswordHashAlgorithm

from .base import PasswordHasher
from .types import HashInfo, PasswordInfo


class PBKDF2_SHA256Hasher(PasswordHasher):
    """PBKDF2-SHA256 password hashing implementation."""

    def generate_new_hash(self, password_info: PasswordInfo) -> HashInfo:
        salt = secrets.token_bytes(password_info.salt_size)
        # Use hashlib.pbkdf2_hmac for standard PBKDF2
        hashed = hashlib.pbkdf2_hmac(
            "sha256", password_info.password.encode("utf-8"), salt, password_info.rounds
        )

        return HashInfo(
            algorithm=PasswordHashAlgorithm.PBKDF2_SHA256,
            rounds=password_info.rounds,
            salt=salt.hex(),
            hash_value=hashed.hex(),
        )

    def verify(self, password: str, hash_info: HashInfo) -> bool:
        if hash_info.algorithm != PasswordHashAlgorithm.PBKDF2_SHA256:
            return False
        if hash_info.rounds is None or hash_info.salt is None:
            return False

        try:
            salt = bytes.fromhex(hash_info.salt)
        except (ValueError, TypeError):
            return False

        # Hash the password with the same salt and rounds
        test_hash = hashlib.pbkdf2_hmac("sha256", password.encode("utf-8"), salt, hash_info.rounds)

        return test_hash.hex() == hash_info.hash_value


class PBKDF2_SHA3_256Hasher(PasswordHasher):
    """PBKDF2-SHA3-256 password hashing implementation."""

    def generate_new_hash(self, password_info: PasswordInfo) -> HashInfo:
        salt = secrets.token_bytes(password_info.salt_size)
        # Use hashlib.pbkdf2_hmac with SHA3-256
        hashed = hashlib.pbkdf2_hmac(
            "sha3_256", password_info.password.encode("utf-8"), salt, password_info.rounds
        )

        return HashInfo(
            algorithm=PasswordHashAlgorithm.PBKDF2_SHA3_256,
            rounds=password_info.rounds,
            salt=salt.hex(),
            hash_value=hashed.hex(),
        )

    def verify(self, password: str, hash_info: HashInfo) -> bool:
        if hash_info.algorithm != PasswordHashAlgorithm.PBKDF2_SHA3_256:
            return False
        if hash_info.rounds is None or hash_info.salt is None:
            return False

        try:
            salt = bytes.fromhex(hash_info.salt)
        except (ValueError, TypeError):
            return False

        # Hash the password with the same salt and rounds
        test_hash = hashlib.pbkdf2_hmac(
            "sha3_256", password.encode("utf-8"), salt, hash_info.rounds
        )

        return test_hash.hex() == hash_info.hash_value



"""Registry for status-based transition hooks.

Maps session statuses to their corresponding hook implementations.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.agent.pool import AgentClientPool
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.repositories.deployment.repository import DeploymentRepository

from .status import (
    RunningHookDependencies,
    RunningTransitionHook,
    StatusTransitionHook,
    TerminatedHookDependencies,
    TerminatedTransitionHook,
)

if TYPE_CHECKING:
    from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository

log = BraceStyleAdapter(logging.getLogger(__name__))


@dataclass
class HookRegistryArgs:
    """Arguments for creating HookRegistry."""

    scheduler_repository: SchedulerRepository
    deployment_repository: DeploymentRepository
    agent_client_pool: AgentClientPool
    event_producer: EventProducer


class HookRegistry:
    """Registry for status-based transition hooks.

    Provides hooks for specific session status transitions.
    Returns None for statuses that don't require hook execution.
    """

    _status_hooks: dict[SessionStatus, StatusTransitionHook]

    def __init__(self, args: HookRegistryArgs) -> None:
        self._status_hooks = {}
        self._initialize_hooks(args)

    def _initialize_hooks(self, args: HookRegistryArgs) -> None:
        """Initialize status-based hooks."""
        # RUNNING transition hook
        running_deps = RunningHookDependencies(
            scheduler_repository=args.scheduler_repository,
            agent_client_pool=args.agent_client_pool,
            deployment_repository=args.deployment_repository,
            event_producer=args.event_producer,
        )
        self._status_hooks[SessionStatus.RUNNING] = RunningTransitionHook(running_deps)

        # TERMINATED transition hook
        terminated_deps = TerminatedHookDependencies(
            deployment_repository=args.deployment_repository,
            event_producer=args.event_producer,
        )
        self._status_hooks[SessionStatus.TERMINATED] = TerminatedTransitionHook(terminated_deps)

    def get_hook(self, status: SessionStatus) -> StatusTransitionHook | None:
        """Get the hook for a specific status transition.

        Args:
            status: The target status of the transition

        Returns:
            The hook for this status, or None if no hook is needed
        """
        hook = self._status_hooks.get(status)
        if hook:
            log.trace("Found hook {} for status {}", hook.__class__.__name__, status)
        return hook



"""User GraphQL query resolvers."""

from __future__ import annotations

from uuid import UUID

import strawberry
from strawberry import Info

from ai.backend.common.contexts.user import current_user
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.user.fetcher import (
    fetch_admin_users,
    fetch_domain_users,
    fetch_project_users,
)
from ai.backend.manager.api.gql.user.types import (
    DomainUserScopeGQL,
    ProjectUserScopeGQL,
    UserFilterGQL,
    UserOrderByGQL,
    UserV2Connection,
    UserV2GQL,
)
from ai.backend.manager.services.user.actions.get_user import GetUserAction


@strawberry.field(
    description=(
        "Added in 26.2.0. Get a single user by UUID (admin only). "
        "Requires superadmin privileges. Returns an error if user is not found."
    )
)  # type: ignore[misc]
async def admin_user_v2(
    info: Info[StrawberryGQLContext],
    user_id: UUID,
) -> UserV2GQL | None:
    """Get a single user by UUID.

    Args:
        info: Strawberry GraphQL context.
        user_id: UUID of the user to retrieve.

    Returns:
        UserV2GQL object.

    Raises:
        UserNotFound: If the user with the given UUID does not exist.
    """
    processors = info.context.processors

    # Execute GetUserAction via processor
    action_result = await processors.user.get_user.wait_for_complete(
        GetUserAction(user_uuid=user_id)
    )

    return UserV2GQL.from_data(action_result.user)


@strawberry.field(
    description=(
        "Added in 26.2.0. List all users with filtering and pagination (admin only). "
        "Requires superadmin privileges."
    )
)  # type: ignore[misc]
async def admin_users_v2(
    info: Info[StrawberryGQLContext],
    filter: UserFilterGQL | None = None,
    order_by: list[UserOrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> UserV2Connection | None:
    """List all users with optional filtering, ordering, and pagination.

    Args:
        info: Strawberry GraphQL context.
        filter: Optional filter criteria.
        order_by: Optional ordering specification.
        before: Cursor for backward pagination.
        after: Cursor for forward pagination.
        first: Number of items from the start.
        last: Number of items from the end.
        limit: Maximum number of items (offset-based).
        offset: Starting position (offset-based).

    Returns:
        UserV2Connection with paginated user records.
    """
    return await fetch_admin_users(
        info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(
    description=(
        "Added in 26.2.0. List users within a specific domain. "
        "Requires domain admin privileges or higher."
    )
)  # type: ignore[misc]
async def domain_users_v2(
    info: Info[StrawberryGQLContext],
    scope: DomainUserScopeGQL,
    filter: UserFilterGQL | None = None,
    order_by: list[UserOrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> UserV2Connection | None:
    """List users within a specific domain.

    Args:
        info: Strawberry GraphQL context.
        scope: Domain scope specifying which domain to query.
        filter: Optional additional filter criteria.
        order_by: Optional ordering specification.
        before: Cursor for backward pagination.
        after: Cursor for forward pagination.
        first: Number of items from the start.
        last: Number of items from the end.
        limit: Maximum number of items (offset-based).
        offset: Starting position (offset-based).

    Returns:
        UserV2Connection with paginated user records from the domain.
    """
    from ai.backend.manager.repositories.user.types import DomainUserSearchScope

    return await fetch_domain_users(
        info,
        scope=DomainUserSearchScope(domain_name=scope.domain_name),
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(
    description=(
        "Added in 26.2.0. List users within a specific project. "
        "Requires project membership or higher privileges."
    )
)  # type: ignore[misc]
async def project_users_v2(
    info: Info[StrawberryGQLContext],
    scope: ProjectUserScopeGQL,
    filter: UserFilterGQL | None = None,
    order_by: list[UserOrderByGQL] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> UserV2Connection | None:
    """List users within a specific project.

    Args:
        info: Strawberry GraphQL context.
        scope: Project scope specifying which project to query.
        filter: Optional additional filter criteria.
        order_by: Optional ordering specification.
        before: Cursor for backward pagination.
        after: Cursor for forward pagination.
        first: Number of items from the start.
        last: Number of items from the end.
        limit: Maximum number of items (offset-based).
        offset: Starting position (offset-based).

    Returns:
        UserV2Connection with paginated user records from the project.
    """
    from ai.backend.manager.repositories.user.types import ProjectUserSearchScope

    return await fetch_project_users(
        info,
        scope=ProjectUserSearchScope(project_id=scope.project_id),
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(
    description=(
        "Added in 26.2.0. Get the current authenticated user's information. "
        "Returns the user associated with the current session. "
        "Returns an error if not authenticated."
    )
)  # type: ignore[misc]
async def my_user_v2(
    info: Info[StrawberryGQLContext],
) -> UserV2GQL | None:
    """Get the current authenticated user's information.

    Args:
        info: Strawberry GraphQL context.

    Returns:
        UserV2GQL for the current user.

    Raises:
        Unauthorized: If the user is not authenticated.
        UserNotFound: If the current user does not exist.
    """
    # Get current authenticated user
    me = current_user()
    if me is None:
        from aiohttp import web

        raise web.HTTPUnauthorized(reason="Authentication required")

    processors = info.context.processors

    # Execute GetUserAction via processor
    action_result = await processors.user.get_user.wait_for_complete(
        GetUserAction(user_uuid=me.user_id)
    )

    return UserV2GQL.from_data(action_result.user)



from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
from typing import Any, cast
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.engine import CursorResult
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import selectinload

from ai.backend.common.data.permission.types import OperationType
from ai.backend.manager.data.permission.id import ObjectId
from ai.backend.manager.models.rbac_models.permission.object_permission import ObjectPermissionRow
from ai.backend.manager.models.rbac_models.role import RoleRow

# =============================================================================
# Data Classes
# =============================================================================


@dataclass
class RBACRevoker:
    """
    Data class for revoking object-level permissions from specific role(s).

    Note: Only entity-level revoking is supported. Field-level revoking is not supported.

    Attributes:
        entity_id: The entity to revoke access from.
        target_role_ids: The role ID(s) to revoke permissions from.
        operations: Operations to revoke. If None, revokes all operations for the entity.
    """

    entity_id: ObjectId
    target_role_ids: list[UUID]
    operations: list[OperationType] | None = None


# =============================================================================
# Query Helpers
# =============================================================================


async def _get_roles_with_permissions(
    db_sess: SASession,
    role_ids: Collection[UUID],
) -> list[RoleRow]:
    """
    Get specified roles with their object permissions.
    """
    if not role_ids:
        return []

    role_scalars = await db_sess.scalars(
        sa.select(RoleRow)
        .where(RoleRow.id.in_(role_ids))
        .options(
            selectinload(RoleRow.object_permission_rows).selectinload(
                ObjectPermissionRow.scope_association_rows
            ),
        )
    )
    return list(role_scalars.unique().all())


# =============================================================================
# ID Collection Helpers (Pure Functions)
# =============================================================================


def _find_object_permissions_to_revoke(
    role_rows: Collection[RoleRow],
    entity_id: ObjectId,
    operations: list[OperationType] | None,
) -> list[UUID]:
    """
    Collect object permission IDs to revoke for the given entity and operations.

    Args:
        role_rows: Roles with loaded object_permissions
        entity_id: Entity to revoke permissions from
        operations: Operations to revoke. If None, revokes all operations.
    """
    object_permission_ids: list[UUID] = []
    for role_row in role_rows:
        for obj_perm in role_row.object_permission_rows:
            if obj_perm.object_id() != entity_id:
                continue
            if operations is None or obj_perm.operation in operations:
                object_permission_ids.append(obj_perm.id)
    return object_permission_ids


# =============================================================================
# Deletion Helpers
# =============================================================================


async def _delete_object_permissions(
    db_sess: SASession,
    ids: Collection[UUID],
) -> int:
    """Delete ObjectPermissionRows by IDs. Returns count of deleted rows."""
    if not ids:
        return 0
    result = await db_sess.execute(
        sa.delete(ObjectPermissionRow).where(ObjectPermissionRow.id.in_(ids))
    )
    return cast(CursorResult[Any], result).rowcount or 0


# =============================================================================
# Public API
# =============================================================================


async def execute_rbac_revoker(
    db_sess: SASession,
    revoker: RBACRevoker,
) -> None:
    """
    Revoke object-level permissions from specified roles.

    This is used when unsharing an entity from specific roles.
    For example, when user A revokes user B's access to a VFolder:
    - User B's role loses object permissions for that VFolder.

    Deletion order:
    1. ObjectPermissionRows - permissions being revoked

    Args:
        db_sess: Async SQLAlchemy session (must be writable).
        revoker: Revoker instance containing entity_id, target_role_ids, and optional operations.
    """
    if not revoker.target_role_ids:
        return

    # Collect related data with eager loading
    role_rows = await _get_roles_with_permissions(
        db_sess,
        revoker.target_role_ids,
    )

    # Find what to delete
    object_permission_ids = _find_object_permissions_to_revoke(
        role_rows,
        revoker.entity_id,
        revoker.operations,
    )

    # Execute deletions
    await _delete_object_permissions(db_sess, object_permission_ids)

    await db_sess.flush()



from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.error_log.types import ErrorLogData, ErrorLogListResult
from ai.backend.manager.models.error_logs import ErrorLogRow
from ai.backend.manager.repositories.base import BatchQuerier, Creator

from .db_source import ErrorLogDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("ErrorLogRepository",)

error_log_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.ERROR_LOG_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ErrorLogRepository:
    _db_source: ErrorLogDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = ErrorLogDBSource(db)

    @error_log_repository_resilience.apply()
    async def create(self, creator: Creator[ErrorLogRow]) -> ErrorLogData:
        return await self._db_source.create(creator)

    @error_log_repository_resilience.apply()
    async def search(
        self,
        querier: BatchQuerier,
    ) -> ErrorLogListResult:
        """Search error logs with querier pattern."""
        return await self._db_source.search(querier=querier)



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.models.scaling_group import ScalingGroupForKeypairsRow
from ai.backend.manager.repositories.base.creator import BulkCreator

from .keypair_base import ScalingGroupKeypairAction


@dataclass
class AssociateScalingGroupWithKeypairsAction(ScalingGroupKeypairAction):
    """Action to associate a scaling group with multiple keypairs."""

    bulk_creator: BulkCreator[ScalingGroupForKeypairsRow]

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE

    @override
    def entity_id(self) -> str | None:
        return None


@dataclass
class AssociateScalingGroupWithKeypairsActionResult(BaseActionResult):
    """Result of associating a scaling group with keypairs."""

    @override
    def entity_id(self) -> str | None:
        return None



import logging
import uuid
from collections.abc import Awaitable, Callable, Sequence
from datetime import UTC, datetime

from ai.backend.common.exception import BackendAIError, ErrorCode
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.actions.action import (
    BaseAction,
    BaseActionResult,
    BaseActionResultMeta,
    BaseActionTriggerMeta,
    ProcessResult,
)
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.types import OperationStatus
from ai.backend.manager.actions.validator.base import ActionValidator

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ActionRunner[TAction: BaseAction, TActionResult: BaseActionResult]:
    _func: Callable[[TAction], Awaitable[TActionResult]]
    _monitors: Sequence[ActionMonitor]

    def __init__(
        self,
        func: Callable[[TAction], Awaitable[TActionResult]],
        monitors: Sequence[ActionMonitor] | None,
    ) -> None:
        self._func = func
        self._monitors = monitors or []

    async def _start_monitors(
        self, action: TAction, action_trigger_meta: BaseActionTriggerMeta
    ) -> None:
        for monitor in self._monitors:
            try:
                await monitor.prepare(action, action_trigger_meta)
            except Exception as e:
                log.warning("Error in monitor prepare method: {}", e)

    async def _finalize_monitors(
        self,
        action: TAction,
        meta: BaseActionResultMeta,
    ) -> None:
        process_result = ProcessResult(meta=meta)
        for monitor in reversed(self._monitors):
            try:
                await monitor.done(action, process_result)
            except Exception as e:
                log.warning("Error in monitor done method: {}", e)

    async def run(
        self, action: TAction, action_trigger_meta: BaseActionTriggerMeta
    ) -> TActionResult:
        started_at = action_trigger_meta.started_at
        action_id = action_trigger_meta.action_id
        status = OperationStatus.UNKNOWN
        description: str = "unknown"
        result: TActionResult | None = None
        error_code: ErrorCode | None = None

        await self._start_monitors(action, action_trigger_meta)
        try:
            result = await self._func(action)
        except BackendAIError as e:
            log.exception("Action processing error: {}", e)
            status = OperationStatus.ERROR
            description = str(e)
            error_code = e.error_code()
            raise
        except BaseException as e:
            log.exception("Unexpected error during action processing: {}", e)
            status = OperationStatus.ERROR
            description = str(e)
            error_code = ErrorCode.default()
            raise
        else:
            status = OperationStatus.SUCCESS
            description = "Success"
            return result
        finally:
            entity_id = action.entity_id()
            if entity_id is None and result is not None:
                entity_id = result.entity_id()
            ended_at = datetime.now(UTC)
            duration = ended_at - started_at
            meta = BaseActionResultMeta(
                action_id=action_id,
                entity_id=entity_id,
                status=status,
                description=description,
                started_at=started_at,
                ended_at=ended_at,
                duration=duration,
                error_code=error_code,
            )
            await self._finalize_monitors(
                action,
                meta,
            )


class ActionProcessor[TAction: BaseAction, TActionResult: BaseActionResult]:
    _validators: Sequence[ActionValidator]

    _runner: ActionRunner[TAction, TActionResult]

    def __init__(
        self,
        func: Callable[[TAction], Awaitable[TActionResult]],
        monitors: Sequence[ActionMonitor] | None = None,
        validators: Sequence[ActionValidator] | None = None,
    ) -> None:
        self._runner = ActionRunner(func, monitors)

        self._validators = validators or []

    async def _run(self, action: TAction) -> TActionResult:
        started_at = datetime.now(UTC)
        action_id = uuid.uuid4()
        action_trigger_meta = BaseActionTriggerMeta(action_id=action_id, started_at=started_at)
        for validator in self._validators:
            await validator.validate(action, action_trigger_meta)

        return await self._runner.run(action, action_trigger_meta)

    async def wait_for_complete(self, action: TAction) -> TActionResult:
        return await self._run(action)



from __future__ import annotations

import base64
import os
import secrets
import uuid
from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Self, TypedDict

import sqlalchemy as sa
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519, rsa
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA256
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship
from sqlalchemy.sql.expression import false

from ai.backend.common import msgpack
from ai.backend.common.types import AccessKey, SecretKey
from ai.backend.manager.data.keypair.types import GeneratedKeyPairData, KeyPairCreator, KeyPairData
from ai.backend.manager.defs import RESERVED_DOTFILES
from ai.backend.manager.models.base import (
    GUID,
    Base,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.resource_policy import KeyPairResourcePolicyRow
    from ai.backend.manager.models.scaling_group import ScalingGroupForKeypairsRow
    from ai.backend.manager.models.session import SessionRow
    from ai.backend.manager.models.user import UserRow

__all__: Sequence[str] = (
    "MAXIMUM_DOTFILE_SIZE",
    "Dotfile",
    "KeyPairRow",
    "keypairs",
    "query_bootstrap_script",
    "query_owned_dotfiles",
    "verify_dotfile_name",
)


MAXIMUM_DOTFILE_SIZE = 64 * 1024  # 61 KiB


# Defined for avoiding circular import
def _get_session_row_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.session import SessionRow

    return KeyPairRow.access_key == foreign(SessionRow.access_key)


class KeyPairRow(Base):  # type: ignore[misc]
    __tablename__ = "keypairs"

    user_id: Mapped[str | None] = mapped_column("user_id", sa.String(length=256), index=True)
    access_key: Mapped[str] = mapped_column("access_key", sa.String(length=20), primary_key=True)
    secret_key: Mapped[str | None] = mapped_column("secret_key", sa.String(length=40))
    is_active: Mapped[bool | None] = mapped_column("is_active", sa.Boolean, index=True)
    is_admin: Mapped[bool | None] = mapped_column(
        "is_admin", sa.Boolean, index=True, default=False, server_default=false()
    )
    created_at: Mapped[datetime | None] = mapped_column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.func.now()
    )
    modified_at: Mapped[datetime | None] = mapped_column(
        "modified_at",
        sa.DateTime(timezone=True),
        server_default=sa.func.now(),
        onupdate=sa.func.current_timestamp(),
    )
    last_used: Mapped[datetime | None] = mapped_column(
        "last_used", sa.DateTime(timezone=True), nullable=True
    )
    rate_limit: Mapped[int | None] = mapped_column("rate_limit", sa.Integer)
    num_queries: Mapped[int | None] = mapped_column("num_queries", sa.Integer, server_default="0")
    # SSH Keypairs.
    ssh_public_key: Mapped[str | None] = mapped_column("ssh_public_key", sa.Text, nullable=True)
    ssh_private_key: Mapped[str | None] = mapped_column("ssh_private_key", sa.Text, nullable=True)
    user: Mapped[uuid.UUID] = mapped_column(
        "user", GUID, sa.ForeignKey("users.uuid"), nullable=False
    )
    resource_policy: Mapped[str] = mapped_column(
        "resource_policy",
        sa.String(length=256),
        sa.ForeignKey("keypair_resource_policies.name"),
        nullable=False,
    )
    # dotfiles column, \x90 means empty list in msgpack
    dotfiles: Mapped[bytes] = mapped_column(
        "dotfiles", sa.LargeBinary(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=b"\x90"
    )
    bootstrap_script: Mapped[str] = mapped_column(
        "bootstrap_script", sa.String(length=MAXIMUM_DOTFILE_SIZE), nullable=False, default=""
    )

    # Relationships
    sessions: Mapped[list[SessionRow]] = relationship(
        "SessionRow",
        primaryjoin=_get_session_row_join_condition,
        foreign_keys="SessionRow.access_key",
        back_populates="access_key_row",
    )
    resource_policy_row: Mapped[KeyPairResourcePolicyRow] = relationship(
        "KeyPairResourcePolicyRow", back_populates="keypairs"
    )
    sgroup_for_keypairs_rows: Mapped[list[ScalingGroupForKeypairsRow]] = relationship(
        "ScalingGroupForKeypairsRow",
        back_populates="keypair_row",
    )
    user_row: Mapped[UserRow] = relationship(
        "UserRow", back_populates="keypairs", foreign_keys=[user]
    )

    @property
    def mapping(self) -> dict[str, object]:
        return {
            "user_id": self.user_id,
            "access_key": self.access_key,
            "secret_key": self.secret_key,
            "is_active": self.is_active,
            "is_admin": self.is_admin,
            "created_at": self.created_at,
            "modified_at": self.modified_at,
            "last_used": self.last_used,
            "rate_limit": self.rate_limit,
            "num_queries": self.num_queries,
            "ssh_public_key": self.ssh_public_key,
            "ssh_private_key": self.ssh_private_key,
            "user": self.user,
            "resource_policy": self.resource_policy,
            "dotfiles": self.dotfiles,
            "bootstrap_script": self.bootstrap_script,
        }

    @classmethod
    def from_creator(
        cls,
        creator: KeyPairCreator,
        generated_data: GeneratedKeyPairData,
        user_id: uuid.UUID,
        email: str,
    ) -> Self:
        return cls(
            user_id=email,
            user=user_id,
            access_key=generated_data.access_key,
            secret_key=generated_data.secret_key,
            is_active=creator.is_active,
            is_admin=creator.is_admin,
            resource_policy=creator.resource_policy,
            rate_limit=creator.rate_limit,
            num_queries=0,
            ssh_public_key=generated_data.ssh_public_key,
            ssh_private_key=generated_data.ssh_private_key,
        )

    def to_data(self) -> KeyPairData:
        if self.secret_key is None:
            raise ValueError("secret_key is required for KeyPairData")
        return KeyPairData(
            user_id=self.user,
            access_key=AccessKey(self.access_key),
            secret_key=SecretKey(self.secret_key),
            is_active=self.is_active if self.is_active is not None else True,
            is_admin=self.is_admin if self.is_admin is not None else False,
            created_at=self.created_at,
            modified_at=self.modified_at,
            resource_policy_name=self.resource_policy,
            rate_limit=self.rate_limit if self.rate_limit is not None else 0,
            ssh_public_key=self.ssh_public_key,
            ssh_private_key=self.ssh_private_key,
            dotfiles=self.dotfiles if self.dotfiles else b"\x90",
            bootstrap_script=self.bootstrap_script,
        )


# NOTE: Deprecated legacy table reference for backward compatibility.
# Use KeyPairRow class directly for new code.
keypairs = KeyPairRow.__table__


class Dotfile(TypedDict):
    data: str
    path: str
    perm: str


def generate_keypair() -> tuple[AccessKey, SecretKey]:
    """
    AWS-like access key and secret key generation.
    """
    ak = "AKIA" + base64.b32encode(secrets.token_bytes(10)).decode("ascii")
    sk = secrets.token_urlsafe(30)
    return AccessKey(ak), SecretKey(sk)


def generate_ssh_keypair() -> tuple[str, str]:
    """
    Generate RSA keypair for SSH/SFTP connection.
    """
    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048,
    )
    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.TraditionalOpenSSL,
        crypto_serialization.NoEncryption(),
    ).decode("utf-8")
    public_key = (
        key.public_key()
        .public_bytes(
            crypto_serialization.Encoding.OpenSSH,
            crypto_serialization.PublicFormat.OpenSSH,
        )
        .decode("utf-8")
    )
    public_key = f"{public_key.rstrip()}\n"
    private_key = f"{private_key.rstrip()}\n"
    return (public_key, private_key)


def prepare_new_keypair(user_email: str, creator: KeyPairCreator) -> dict[str, object]:
    ak, sk = generate_keypair()
    pubkey, privkey = generate_ssh_keypair()
    return {
        "user_id": user_email,
        "access_key": ak,
        "secret_key": sk,
        "is_active": creator.is_active,
        "is_admin": creator.is_admin,
        "resource_policy": creator.resource_policy,
        "rate_limit": creator.rate_limit,
        "num_queries": 0,
        "ssh_public_key": pubkey,
        "ssh_private_key": privkey,
    }


def generate_keypair_data() -> GeneratedKeyPairData:
    ak, sk = generate_keypair()
    pubkey, privkey = generate_ssh_keypair()
    return GeneratedKeyPairData(
        access_key=ak,
        secret_key=sk,
        ssh_public_key=pubkey,
        ssh_private_key=privkey,
    )


def _generate_random_bytes_to_verify_keypairs() -> bytes:
    # Check if the keys match by signing and verifying a test message
    return os.urandom(32)


def _check_rsa_keypair(private_key: rsa.RSAPrivateKey, public_key: rsa.RSAPublicKey) -> None:
    test_message = _generate_random_bytes_to_verify_keypairs()
    signature = private_key.sign(test_message, PKCS1v15(), SHA256())
    public_key.verify(signature, test_message, PKCS1v15(), SHA256())


def _check_ecdsa_keypair(
    private_key: ec.EllipticCurvePrivateKey, public_key: ec.EllipticCurvePublicKey
) -> None:
    test_message = _generate_random_bytes_to_verify_keypairs()
    signature = private_key.sign(test_message, ec.ECDSA(SHA256()))
    public_key.verify(signature, test_message, ec.ECDSA(SHA256()))


def _check_ed25519_keypair(
    private_key: ed25519.Ed25519PrivateKey, public_key: ed25519.Ed25519PublicKey
) -> None:
    test_message = _generate_random_bytes_to_verify_keypairs()
    signature = private_key.sign(test_message)
    public_key.verify(signature, test_message)


def validate_ssh_keypair(private_key_value: str, public_key_value: str) -> tuple[bool, str | None]:
    """
    Validate RSA keypair for SSH/SFTP connection.
    Args:
        private_key_value: PEM-encoded private key string (OpenSSL format)
        public_key_value: OpenSSH-encoded public key string
    Returns:
        tuple[bool, Optional[str]]:
            Tuple containing a boolean indicating if the keypair is valid,
            and an optional error message if invalid.
    """

    try:
        # Load the private key (PEM format)
        private_key = crypto_serialization.load_pem_private_key(
            private_key_value.encode(),
            password=None,  # No encryption as specified
        )
    except ValueError:
        return False, "Invalid private key format"

    try:
        # Load the public key (OpenSSH format)
        public_key = crypto_serialization.load_ssh_public_key(public_key_value.encode())
    except ValueError:
        return False, "Invalid public key format"

    try:
        match private_key, public_key:
            case rsa.RSAPrivateKey(), rsa.RSAPublicKey():
                _check_rsa_keypair(private_key, public_key)
            case ec.EllipticCurvePrivateKey(), ec.EllipticCurvePublicKey():
                _check_ecdsa_keypair(private_key, public_key)
            case ed25519.Ed25519PrivateKey(), ed25519.Ed25519PublicKey():
                _check_ed25519_keypair(private_key, public_key)
            case _:
                return (
                    False,
                    f"Unsupported pair of keys: private={type(private_key)}, public={type(public_key)}",
                )
    except InvalidSignature as e:
        return False, f"Keypair does not match: {e}"

    return True, None


async def query_owned_dotfiles(
    conn: SAConnection,
    access_key: AccessKey,
) -> tuple[list[Dotfile], int]:
    query = (
        sa.select(KeyPairRow.dotfiles)
        .select_from(KeyPairRow)
        .where(KeyPairRow.access_key == access_key)
    )
    packed_dotfile = (await conn.execute(query)).scalar()
    if packed_dotfile is None:
        return [], MAXIMUM_DOTFILE_SIZE
    rows = msgpack.unpackb(packed_dotfile)
    return rows, MAXIMUM_DOTFILE_SIZE - len(packed_dotfile)


async def query_bootstrap_script(
    conn: SAConnection,
    access_key: AccessKey,
) -> tuple[str, int]:
    query = (
        sa.select(KeyPairRow.bootstrap_script)
        .select_from(KeyPairRow)
        .where(KeyPairRow.access_key == access_key)
    )
    script = (await conn.execute(query)).scalar()
    if script is None:
        return "", MAXIMUM_DOTFILE_SIZE
    return script, MAXIMUM_DOTFILE_SIZE - len(script)


def verify_dotfile_name(dotfile: str) -> bool:
    return dotfile not in RESERVED_DOTFILES



"""Action for destroying deployments."""

import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction


@dataclass
class DestroyDeploymentAction(DeploymentBaseAction):
    """Action to destroy an existing deployment."""

    endpoint_id: uuid.UUID

    @override
    def entity_id(self) -> str | None:
        return str(self.endpoint_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DestroyDeploymentActionResult(BaseActionResult):
    success: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""Fair Share Row models.

Database models for Fair Share state management per resource group:
- DomainFairShareRow: Domain-level fair share state
- ProjectFairShareRow: Project-level fair share state
- UserFairShareRow: User-level fair share state (per project)
"""

from __future__ import annotations

import uuid
from datetime import date, datetime
from decimal import Decimal
from typing import TYPE_CHECKING

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.data.fair_share import (
    DomainFairShareData,
    FairShareCalculationSnapshot,
    FairShareData,
    FairShareMetadata,
    FairShareSpec,
    ProjectFairShareData,
    UserFairShareData,
)
from ai.backend.manager.models.base import (
    GUID,
    Base,
    ResourceSlotColumn,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.domain import DomainRow
    from ai.backend.manager.models.group import GroupRow
    from ai.backend.manager.models.user import UserRow


__all__ = (
    "DomainFairShareRow",
    "ProjectFairShareRow",
    "UserFairShareRow",
)

# Default values for Fair Share calculation parameters
DEFAULT_HALF_LIFE_DAYS = 7
DEFAULT_LOOKBACK_DAYS = 28
DEFAULT_DECAY_UNIT_DAYS = 1


def _merge_resource_weights(
    row_weights: ResourceSlot,
    default_weight: Decimal,
    available_slots: ResourceSlot,
) -> tuple[ResourceSlot, frozenset[str]]:
    """Merge explicit and default resource weights.

    Args:
        row_weights: Entity's configured weights (may be incomplete)
        default_weight: Scaling group's default weight (single value for all resources)
        available_slots: Available resource types in cluster

    Returns:
        Tuple of (merged_weights, uses_default_resources)
        - merged_weights: Complete ResourceSlot with all available resource types
        - uses_default_resources: Set of resource types using default weight
    """
    merged = {}
    uses_default = []

    for resource_type in available_slots:
        if resource_type in row_weights:
            # Use explicit weight
            merged[resource_type] = row_weights[resource_type]
        else:
            # Use default_weight for all missing resources
            merged[resource_type] = default_weight
            uses_default.append(resource_type)

    return ResourceSlot(merged), frozenset(uses_default)


def _get_domain_fair_share_domain_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.domain import DomainRow

    return DomainFairShareRow.domain_name == foreign(DomainRow.name)


class DomainFairShareRow(Base):  # type: ignore[misc]
    """Per-domain Fair Share state.

    Stores weight (configured value) and calculated values together for current state.
    One row per (resource_group, domain_name) combination.
    """

    __tablename__ = "domain_fair_shares"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    resource_group: Mapped[str] = mapped_column(
        "resource_group", sa.String(length=64), nullable=False, index=True
    )
    domain_name: Mapped[str] = mapped_column(
        "domain_name", sa.String(length=64), nullable=False, index=True
    )

    # Numeric(precision=10, scale=4): Total 10 digits with 4 decimal places.
    # Range: -999999.9999 ~ 999999.9999
    # Example: weight=1.5 means 1.5x priority multiplier compared to default weight=1.0
    # When NULL, the resource group's default_weight should be used.
    weight: Mapped[Decimal | None] = mapped_column(
        "weight",
        sa.Numeric(precision=10, scale=4),
        nullable=True,
        default=None,
        comment="Priority weight multiplier. Higher weight = higher priority allocation ratio. "
        "Example: 2.0 means 2x priority compared to default 1.0. "
        "NULL means use resource group's default_weight.",
    )

    # Numeric(precision=8, scale=6): Total 8 digits with 6 decimal places.
    # Range: 0.000000 ~ 99.999999 (typically 0.0 ~ 1.0)
    # Calculated using formula: F = 2^(-normalized_usage / weight)
    fair_share_factor: Mapped[Decimal] = mapped_column(
        "fair_share_factor",
        sa.Numeric(precision=8, scale=6),
        nullable=False,
        default=Decimal("1.0"),
        comment="Calculated priority score from 0.0 to 1.0. "
        "Higher value = less past usage = higher scheduling priority. "
        "Formula: F = 2^(-normalized_usage / weight)",
    )
    total_decayed_usage: Mapped[ResourceSlot] = mapped_column(
        "total_decayed_usage",
        ResourceSlotColumn(),
        nullable=False,
        default=ResourceSlot,
        comment="Sum of historical resource usage with time decay applied. "
        "Unit: resource-seconds (e.g., cpu-seconds, mem-seconds). "
        "Older usage is weighted less via half-life decay.",
    )
    resource_weights: Mapped[ResourceSlot] = mapped_column(
        "resource_weights",
        ResourceSlotColumn(),
        nullable=False,
        default=ResourceSlot,
        comment="Resource weights used in fair share calculation. "
        "From scheduler_opts. Example: {cpu: 1.0, mem: 1.0, cuda.device: 10.0}",
    )
    # Numeric(precision=8, scale=6): Total 8 digits with 6 decimal places.
    # Range: 0.000000 ~ 99.999999 (typically 0.0 ~ 1.0)
    # IMPORTANT: Both usage and capacity must be in resource-seconds for correct ratio.
    # - usage: resource-seconds (e.g., 3600 GPU-seconds = 1 GPU-hour)
    # - capacity: resource-seconds (e.g., 8 GPUs × 86400 sec/day × 28 days)
    # DO NOT mix resource-seconds (usage) with raw resource units (capacity).
    normalized_usage: Mapped[Decimal] = mapped_column(
        "normalized_usage",
        sa.Numeric(precision=8, scale=6),
        nullable=False,
        default=Decimal("0"),
        comment="Weighted average of (usage/capacity) per resource (0.0 ~ 1.0). "
        "Both usage and capacity are in resource-seconds. "
        "Formula: sum((usage[r]/capacity[r]) * weight[r]) / sum(weight[r])",
    )
    last_calculated_at: Mapped[datetime] = mapped_column(
        "last_calculated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        comment="Timestamp when fair_share_factor was last recalculated by batch job.",
    )

    lookback_start: Mapped[date] = mapped_column(
        "lookback_start",
        sa.Date,
        nullable=False,
        server_default=sa.func.current_date(),
        comment="Start date of the usage history window used in last calculation. "
        "Typically: today - lookback_days.",
    )
    lookback_end: Mapped[date] = mapped_column(
        "lookback_end",
        sa.Date,
        nullable=False,
        server_default=sa.func.current_date(),
        comment="End date of the usage history window used in last calculation. Typically: today.",
    )
    half_life_days: Mapped[int] = mapped_column(
        "half_life_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_HALF_LIFE_DAYS,
        comment="Number of days for usage to decay to 50%%. "
        "Example: 7 means usage from 7 days ago counts as 50%% of recent usage.",
    )
    lookback_days: Mapped[int] = mapped_column(
        "lookback_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_LOOKBACK_DAYS,
        comment="Number of days of usage history to consider for fair share calculation. "
        "Example: 28 means only last 28 days of usage affects priority.",
    )
    decay_unit_days: Mapped[int] = mapped_column(
        "decay_unit_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_DECAY_UNIT_DAYS,
        comment="Aggregation period for usage buckets in days. "
        "Example: 1 means daily aggregation, 7 means weekly aggregation.",
    )

    created_at: Mapped[datetime] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )
    updated_at: Mapped[datetime] = mapped_column(
        "updated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        onupdate=sa.func.now(),
    )

    domain: Mapped[DomainRow | None] = relationship(
        "DomainRow",
        primaryjoin=_get_domain_fair_share_domain_join_condition,
        foreign_keys=[domain_name],
        uselist=False,
        viewonly=True,
    )

    __table_args__ = (
        sa.UniqueConstraint("resource_group", "domain_name", name="uq_domain_fair_share"),
        sa.Index("ix_domain_fair_share_lookup", "resource_group", "domain_name"),
    )

    def to_data(
        self,
        default_weight: Decimal,
        available_slots: ResourceSlot,
    ) -> DomainFairShareData:
        """Convert to DomainFairShareData with merged resource weights.

        Args:
            default_weight: Scaling group's default weight (used when self.weight is NULL
                           and as fallback for missing resource weights)
            available_slots: Available resource types in the cluster

        Note: self.weight can be NULL in DB.
        - NULL means "use default weight" (use_default=True)
        - NOT NULL means "explicit weight" (use_default=False)
        """
        # Handle entity weight (existing logic)
        if self.weight is None:
            weight = default_weight
            use_default = True
        else:
            weight = self.weight
            use_default = False

        # Merge resource weights (new)
        merged_weights, uses_default_resources = _merge_resource_weights(
            self.resource_weights,
            default_weight,
            available_slots,
        )

        return DomainFairShareData(
            resource_group=self.resource_group,
            domain_name=self.domain_name,
            data=FairShareData(
                spec=FairShareSpec(
                    weight=weight,
                    half_life_days=self.half_life_days,
                    lookback_days=self.lookback_days,
                    decay_unit_days=self.decay_unit_days,
                    resource_weights=merged_weights,  # Use merged weights
                ),
                calculation_snapshot=FairShareCalculationSnapshot(
                    fair_share_factor=self.fair_share_factor,
                    total_decayed_usage=self.total_decayed_usage,
                    normalized_usage=self.normalized_usage,
                    lookback_start=self.lookback_start,
                    lookback_end=self.lookback_end,
                    last_calculated_at=self.last_calculated_at,
                ),
                metadata=FairShareMetadata(
                    created_at=self.created_at,
                    updated_at=self.updated_at,
                ),
                use_default=use_default,  # True if weight was NULL
                uses_default_resources=uses_default_resources,  # New field
            ),
        )


def _get_project_fair_share_project_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.group import GroupRow

    return ProjectFairShareRow.project_id == foreign(GroupRow.id)


def _get_project_fair_share_domain_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.domain import DomainRow

    return ProjectFairShareRow.domain_name == foreign(DomainRow.name)


class ProjectFairShareRow(Base):  # type: ignore[misc]
    """Per-project Fair Share state.

    One row per (resource_group, project_id) combination.
    """

    __tablename__ = "project_fair_shares"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    resource_group: Mapped[str] = mapped_column(
        "resource_group", sa.String(length=64), nullable=False, index=True
    )
    project_id: Mapped[uuid.UUID] = mapped_column("project_id", GUID, nullable=False, index=True)
    domain_name: Mapped[str] = mapped_column(
        "domain_name", sa.String(length=64), nullable=False, index=True
    )

    weight: Mapped[Decimal | None] = mapped_column(
        "weight",
        sa.Numeric(precision=10, scale=4),
        nullable=True,
        default=None,
        comment="Priority weight multiplier. Higher weight = higher priority allocation ratio. "
        "Example: 2.0 means 2x priority compared to default 1.0. "
        "NULL means use resource group's default_weight.",
    )

    fair_share_factor: Mapped[Decimal] = mapped_column(
        "fair_share_factor",
        sa.Numeric(precision=8, scale=6),
        nullable=False,
        default=Decimal("1.0"),
        comment="Calculated priority score from 0.0 to 1.0. "
        "Higher value = less past usage = higher scheduling priority. "
        "Formula: F = 2^(-normalized_usage / weight)",
    )
    total_decayed_usage: Mapped[ResourceSlot] = mapped_column(
        "total_decayed_usage",
        ResourceSlotColumn(),
        nullable=False,
        default=ResourceSlot,
        comment="Sum of historical resource usage with time decay applied. Unit: resource-seconds.",
    )
    resource_weights: Mapped[ResourceSlot] = mapped_column(
        "resource_weights",
        ResourceSlotColumn(),
        nullable=False,
        default=ResourceSlot,
        comment="Resource weights used in fair share calculation. "
        "From scheduler_opts. Example: {cpu: 1.0, mem: 1.0, cuda.device: 10.0}",
    )
    # IMPORTANT: Both usage and capacity must be in resource-seconds for correct ratio.
    # - usage: resource-seconds (e.g., 3600 GPU-seconds = 1 GPU-hour)
    # - capacity: resource-seconds (e.g., 8 GPUs × 86400 sec/day × 28 days)
    # DO NOT mix resource-seconds (usage) with raw resource units (capacity).
    normalized_usage: Mapped[Decimal] = mapped_column(
        "normalized_usage",
        sa.Numeric(precision=8, scale=6),
        nullable=False,
        default=Decimal("0"),
        comment="Weighted average of (usage/capacity) per resource (0.0 ~ 1.0). "
        "Both usage and capacity are in resource-seconds. "
        "Formula: sum((usage[r]/capacity[r]) * weight[r]) / sum(weight[r])",
    )
    last_calculated_at: Mapped[datetime] = mapped_column(
        "last_calculated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        comment="Timestamp when fair_share_factor was last recalculated.",
    )

    lookback_start: Mapped[date] = mapped_column(
        "lookback_start",
        sa.Date,
        nullable=False,
        server_default=sa.func.current_date(),
        comment="Start date of the usage history window used in last calculation.",
    )
    lookback_end: Mapped[date] = mapped_column(
        "lookback_end",
        sa.Date,
        nullable=False,
        server_default=sa.func.current_date(),
        comment="End date of the usage history window used in last calculation.",
    )
    half_life_days: Mapped[int] = mapped_column(
        "half_life_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_HALF_LIFE_DAYS,
        comment="Number of days for usage to decay to 50%%.",
    )
    lookback_days: Mapped[int] = mapped_column(
        "lookback_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_LOOKBACK_DAYS,
        comment="Number of days of usage history to consider.",
    )
    decay_unit_days: Mapped[int] = mapped_column(
        "decay_unit_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_DECAY_UNIT_DAYS,
        comment="Aggregation period for usage buckets in days.",
    )

    created_at: Mapped[datetime] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )
    updated_at: Mapped[datetime] = mapped_column(
        "updated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        onupdate=sa.func.now(),
    )

    project: Mapped[GroupRow | None] = relationship(
        "GroupRow",
        primaryjoin=_get_project_fair_share_project_join_condition,
        foreign_keys=[project_id],
        uselist=False,
        viewonly=True,
    )
    domain: Mapped[DomainRow | None] = relationship(
        "DomainRow",
        primaryjoin=_get_project_fair_share_domain_join_condition,
        foreign_keys=[domain_name],
        uselist=False,
        viewonly=True,
    )

    __table_args__ = (
        sa.UniqueConstraint("resource_group", "project_id", name="uq_project_fair_share"),
        sa.Index("ix_project_fair_share_lookup", "resource_group", "project_id"),
    )

    def to_data(
        self,
        default_weight: Decimal,
        available_slots: ResourceSlot,
    ) -> ProjectFairShareData:
        """Convert to ProjectFairShareData with merged resource weights.

        Args:
            default_weight: Scaling group's default weight (used when self.weight is NULL
                           and as fallback for missing resource weights)
            available_slots: Available resource types in the cluster

        Note: self.weight can be NULL in DB.
        - NULL means "use default weight" (use_default=True)
        - NOT NULL means "explicit weight" (use_default=False)
        """
        # Handle entity weight (existing logic)
        if self.weight is None:
            weight = default_weight
            use_default = True
        else:
            weight = self.weight
            use_default = False

        # Merge resource weights (new)
        merged_weights, uses_default_resources = _merge_resource_weights(
            self.resource_weights,
            default_weight,
            available_slots,
        )

        return ProjectFairShareData(
            resource_group=self.resource_group,
            project_id=self.project_id,
            domain_name=self.domain_name,
            data=FairShareData(
                spec=FairShareSpec(
                    weight=weight,
                    half_life_days=self.half_life_days,
                    lookback_days=self.lookback_days,
                    decay_unit_days=self.decay_unit_days,
                    resource_weights=merged_weights,  # Use merged weights
                ),
                calculation_snapshot=FairShareCalculationSnapshot(
                    fair_share_factor=self.fair_share_factor,
                    total_decayed_usage=self.total_decayed_usage,
                    normalized_usage=self.normalized_usage,
                    lookback_start=self.lookback_start,
                    lookback_end=self.lookback_end,
                    last_calculated_at=self.last_calculated_at,
                ),
                metadata=FairShareMetadata(
                    created_at=self.created_at,
                    updated_at=self.updated_at,
                ),
                use_default=use_default,  # True if weight was NULL
                uses_default_resources=uses_default_resources,  # New field
            ),
        )


def _get_user_fair_share_user_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.user import UserRow

    return UserFairShareRow.user_uuid == foreign(UserRow.uuid)


def _get_user_fair_share_project_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.group import GroupRow

    return UserFairShareRow.project_id == foreign(GroupRow.id)


def _get_user_fair_share_domain_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.domain import DomainRow

    return UserFairShareRow.domain_name == foreign(DomainRow.name)


class UserFairShareRow(Base):  # type: ignore[misc]
    """Per-user Fair Share state.

    Since a User can belong to multiple Projects, distinguished by
    (user_uuid, project_id) combination.

    One row per (resource_group, user_uuid, project_id) combination.
    """

    __tablename__ = "user_fair_shares"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    resource_group: Mapped[str] = mapped_column(
        "resource_group", sa.String(length=64), nullable=False, index=True
    )
    user_uuid: Mapped[uuid.UUID] = mapped_column("user_uuid", GUID, nullable=False, index=True)
    project_id: Mapped[uuid.UUID] = mapped_column("project_id", GUID, nullable=False, index=True)
    domain_name: Mapped[str] = mapped_column(
        "domain_name", sa.String(length=64), nullable=False, index=True
    )

    weight: Mapped[Decimal | None] = mapped_column(
        "weight",
        sa.Numeric(precision=10, scale=4),
        nullable=True,
        default=None,
        comment="Priority weight multiplier. Higher weight = higher priority allocation ratio. "
        "Example: 2.0 means 2x priority compared to default 1.0. "
        "NULL means use resource group's default_weight.",
    )

    fair_share_factor: Mapped[Decimal] = mapped_column(
        "fair_share_factor",
        sa.Numeric(precision=8, scale=6),
        nullable=False,
        default=Decimal("1.0"),
        comment="Calculated priority score from 0.0 to 1.0. "
        "Higher value = less past usage = higher scheduling priority. "
        "Formula: F = 2^(-normalized_usage / weight)",
    )
    scheduling_rank: Mapped[int | None] = mapped_column(
        "scheduling_rank",
        sa.Integer,
        nullable=True,
        comment="Computed scheduling priority rank. "
        "Lower value = higher priority (1 = highest). "
        "NULL means rank calculation has not been performed yet.",
    )
    total_decayed_usage: Mapped[ResourceSlot] = mapped_column(
        "total_decayed_usage",
        ResourceSlotColumn(),
        nullable=False,
        default=ResourceSlot,
        comment="Sum of historical resource usage with time decay applied. Unit: resource-seconds.",
    )
    resource_weights: Mapped[ResourceSlot] = mapped_column(
        "resource_weights",
        ResourceSlotColumn(),
        nullable=False,
        default=ResourceSlot,
        comment="Resource weights used in fair share calculation. "
        "From scheduler_opts. Example: {cpu: 1.0, mem: 1.0, cuda.device: 10.0}",
    )
    # IMPORTANT: Both usage and capacity must be in resource-seconds for correct ratio.
    # - usage: resource-seconds (e.g., 3600 GPU-seconds = 1 GPU-hour)
    # - capacity: resource-seconds (e.g., 8 GPUs × 86400 sec/day × 28 days)
    # DO NOT mix resource-seconds (usage) with raw resource units (capacity).
    normalized_usage: Mapped[Decimal] = mapped_column(
        "normalized_usage",
        sa.Numeric(precision=8, scale=6),
        nullable=False,
        default=Decimal("0"),
        comment="Weighted average of (usage/capacity) per resource (0.0 ~ 1.0). "
        "Both usage and capacity are in resource-seconds. "
        "Formula: sum((usage[r]/capacity[r]) * weight[r]) / sum(weight[r])",
    )
    last_calculated_at: Mapped[datetime] = mapped_column(
        "last_calculated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        comment="Timestamp when fair_share_factor was last recalculated.",
    )

    lookback_start: Mapped[date] = mapped_column(
        "lookback_start",
        sa.Date,
        nullable=False,
        server_default=sa.func.current_date(),
        comment="Start date of the usage history window used in last calculation.",
    )
    lookback_end: Mapped[date] = mapped_column(
        "lookback_end",
        sa.Date,
        nullable=False,
        server_default=sa.func.current_date(),
        comment="End date of the usage history window used in last calculation.",
    )
    half_life_days: Mapped[int] = mapped_column(
        "half_life_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_HALF_LIFE_DAYS,
        comment="Number of days for usage to decay to 50%%.",
    )
    lookback_days: Mapped[int] = mapped_column(
        "lookback_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_LOOKBACK_DAYS,
        comment="Number of days of usage history to consider.",
    )
    decay_unit_days: Mapped[int] = mapped_column(
        "decay_unit_days",
        sa.Integer,
        nullable=False,
        default=DEFAULT_DECAY_UNIT_DAYS,
        comment="Aggregation period for usage buckets in days.",
    )

    created_at: Mapped[datetime] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
    )
    updated_at: Mapped[datetime] = mapped_column(
        "updated_at",
        sa.DateTime(timezone=True),
        nullable=False,
        server_default=sa.func.now(),
        onupdate=sa.func.now(),
    )

    user: Mapped[UserRow | None] = relationship(
        "UserRow",
        primaryjoin=_get_user_fair_share_user_join_condition,
        foreign_keys=[user_uuid],
        uselist=False,
        viewonly=True,
    )
    project: Mapped[GroupRow | None] = relationship(
        "GroupRow",
        primaryjoin=_get_user_fair_share_project_join_condition,
        foreign_keys=[project_id],
        uselist=False,
        viewonly=True,
    )
    domain: Mapped[DomainRow | None] = relationship(
        "DomainRow",
        primaryjoin=_get_user_fair_share_domain_join_condition,
        foreign_keys=[domain_name],
        uselist=False,
        viewonly=True,
    )

    __table_args__ = (
        sa.UniqueConstraint(
            "resource_group",
            "user_uuid",
            "project_id",
            name="uq_user_fair_share",
        ),
        sa.Index("ix_user_fair_share_lookup", "resource_group", "user_uuid", "project_id"),
    )

    def to_data(
        self,
        default_weight: Decimal,
        available_slots: ResourceSlot,
    ) -> UserFairShareData:
        """Convert to UserFairShareData with merged resource weights.

        Args:
            default_weight: Scaling group's default weight (used when self.weight is NULL
                           and as fallback for missing resource weights)
            available_slots: Available resource types in the cluster

        Note: self.weight can be NULL in DB.
        - NULL means "use default weight" (use_default=True)
        - NOT NULL means "explicit weight" (use_default=False)
        """
        # Handle entity weight (existing logic)
        if self.weight is None:
            weight = default_weight
            use_default = True
        else:
            weight = self.weight
            use_default = False

        # Merge resource weights (new)
        merged_weights, uses_default_resources = _merge_resource_weights(
            self.resource_weights,
            default_weight,
            available_slots,
        )

        return UserFairShareData(
            resource_group=self.resource_group,
            user_uuid=self.user_uuid,
            project_id=self.project_id,
            domain_name=self.domain_name,
            data=FairShareData(
                spec=FairShareSpec(
                    weight=weight,
                    half_life_days=self.half_life_days,
                    lookback_days=self.lookback_days,
                    decay_unit_days=self.decay_unit_days,
                    resource_weights=merged_weights,  # Use merged weights
                ),
                calculation_snapshot=FairShareCalculationSnapshot(
                    fair_share_factor=self.fair_share_factor,
                    total_decayed_usage=self.total_decayed_usage,
                    normalized_usage=self.normalized_usage,
                    lookback_start=self.lookback_start,
                    lookback_end=self.lookback_end,
                    last_calculated_at=self.last_calculated_at,
                ),
                metadata=FairShareMetadata(
                    created_at=self.created_at,
                    updated_at=self.updated_at,
                ),
                use_default=use_default,  # True if weight was NULL
                uses_default_resources=uses_default_resources,  # New field
            ),
            scheduling_rank=self.scheduling_rank,
        )



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class GetRoleAction(AuthAction):
    user_id: uuid.UUID
    group_id: uuid.UUID | None
    is_superadmin: bool
    is_admin: bool

    @override
    def entity_id(self) -> str | None:
        return str(self.user_id)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class GetRoleActionResult(BaseActionResult):
    global_role: str
    domain_role: str
    group_role: str | None

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

from .actions import (
    SearchDeploymentHistoryAction,
    SearchDeploymentHistoryActionResult,
    SearchRouteHistoryAction,
    SearchRouteHistoryActionResult,
    SearchSessionHistoryAction,
    SearchSessionHistoryActionResult,
)
from .processors import SchedulingHistoryProcessors
from .service import SchedulingHistoryService

__all__ = (
    "SchedulingHistoryProcessors",
    "SchedulingHistoryService",
    "SearchSessionHistoryAction",
    "SearchSessionHistoryActionResult",
    "SearchDeploymentHistoryAction",
    "SearchDeploymentHistoryActionResult",
    "SearchRouteHistoryAction",
    "SearchRouteHistoryActionResult",
)



"""Handler for deprioritizing sessions that exceeded max scheduling retries."""

from __future__ import annotations

import logging
from collections.abc import Sequence

from ai.backend.common.defs.session import SESSION_PRIORITY_MIN
from ai.backend.common.types import AccessKey
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus, StatusTransitions, TransitionStatus
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
from ai.backend.manager.sokovan.data import SessionWithKernels
from ai.backend.manager.sokovan.scheduler.handlers.base import SessionLifecycleHandler
from ai.backend.manager.sokovan.scheduler.results import (
    SessionExecutionResult,
    SessionTransitionInfo,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

# Amount to lower priority when deprioritizing
DEPRIORITIZE_AMOUNT = 10


class DeprioritizeSessionsLifecycleHandler(SessionLifecycleHandler):
    """Handler for deprioritizing sessions that exceeded max scheduling retries.

    When sessions in PENDING status exceed max scheduling retries (give_up),
    they transition to DEPRIORITIZING. This handler:
    1. Lowers the session priority
    2. Transitions back to PENDING for re-scheduling with lower priority

    This allows sessions to eventually be scheduled when resources become available,
    rather than being terminated outright.
    """

    def __init__(
        self,
        repository: SchedulerRepository,
    ) -> None:
        self._repository = repository

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "deprioritize-sessions"

    @classmethod
    def target_statuses(cls) -> list[SessionStatus]:
        """Sessions in DEPRIORITIZING state."""
        return [SessionStatus.DEPRIORITIZING]

    @classmethod
    def target_kernel_statuses(cls) -> list[KernelStatus] | None:
        """No kernel filtering for deprioritize."""
        return None

    @classmethod
    def status_transitions(cls) -> StatusTransitions:
        """Define state transitions for deprioritize handler (BEP-1030).

        - success: Session → PENDING (re-schedule with lower priority)
        - need_retry: None
        - expired: None
        - give_up: None
        """
        return StatusTransitions(
            success=TransitionStatus(
                session=SessionStatus.PENDING,
                kernel=None,  # Kernel status unchanged
            ),
            need_retry=None,
            expired=None,
            give_up=None,
        )

    @property
    def lock_id(self) -> LockID | None:
        """No lock needed for deprioritizing sessions."""
        return None

    async def execute(
        self,
        scaling_group: str,
        sessions: Sequence[SessionWithKernels],
    ) -> SessionExecutionResult:
        """Lower priority and prepare for re-scheduling.

        The coordinator provides SessionWithKernels.
        This handler:
        1. Lowers priority for each session via repository
        2. Reports all as success for Coordinator to transition to PENDING
        """
        result = SessionExecutionResult()

        if not sessions:
            return result

        # Lower priority for all sessions (with floor at SESSION_PRIORITY_MIN)
        session_ids = [s.session_info.identity.id for s in sessions]
        await self._repository.lower_session_priority(
            session_ids, DEPRIORITIZE_AMOUNT, SESSION_PRIORITY_MIN
        )

        log.info(
            "Lowered priority by {} for {} sessions in scaling group {}",
            DEPRIORITIZE_AMOUNT,
            len(sessions),
            scaling_group,
        )

        # Mark all sessions as success for status transition to PENDING
        for session in sessions:
            session_info = session.session_info
            result.successes.append(
                SessionTransitionInfo(
                    session_id=session_info.identity.id,
                    from_status=session_info.lifecycle.status,
                    reason="deprioritized-for-rescheduling",
                    creation_id=session_info.identity.creation_id,
                    access_key=AccessKey(session_info.metadata.access_key),
                )
            )

        return result



from __future__ import annotations

import uuid
from collections.abc import Collection

import sqlalchemy as sa

from ai.backend.manager.models.object_storage import ObjectStorageRow
from ai.backend.manager.repositories.base import QueryCondition


class ObjectStorageConditions:
    """Query conditions for object storages."""

    @staticmethod
    def by_ids(storage_ids: Collection[uuid.UUID]) -> QueryCondition:
        def inner() -> sa.sql.expression.ColumnElement[bool]:
            return ObjectStorageRow.id.in_(storage_ids)

        return inner



from .create import CreateHuggingFaceRegistryAction, CreateHuggingFaceRegistryActionResult
from .delete import DeleteHuggingFaceRegistryAction, DeleteHuggingFaceRegistryActionResult
from .get import GetHuggingFaceRegistryAction, GetHuggingFaceRegistryActionResult
from .get_multi import GetHuggingFaceRegistriesAction, GetHuggingFaceRegistriesActionResult
from .list import ListHuggingFaceRegistryAction, ListHuggingFaceRegistryActionResult
from .search import SearchHuggingFaceRegistriesAction, SearchHuggingFaceRegistriesActionResult
from .update import UpdateHuggingFaceRegistryAction, UpdateHuggingFaceRegistryActionResult

__all__ = [
    "CreateHuggingFaceRegistryAction",
    "CreateHuggingFaceRegistryActionResult",
    "DeleteHuggingFaceRegistryAction",
    "DeleteHuggingFaceRegistryActionResult",
    "GetHuggingFaceRegistriesAction",
    "GetHuggingFaceRegistriesActionResult",
    "GetHuggingFaceRegistryAction",
    "GetHuggingFaceRegistryActionResult",
    "ListHuggingFaceRegistryAction",
    "ListHuggingFaceRegistryActionResult",
    "SearchHuggingFaceRegistriesAction",
    "SearchHuggingFaceRegistriesActionResult",
    "UpdateHuggingFaceRegistryAction",
    "UpdateHuggingFaceRegistryActionResult",
]



from __future__ import annotations

import uuid
from datetime import datetime
from typing import (
    TYPE_CHECKING,
)

import sqlalchemy as sa
from sqlalchemy.orm import (
    Mapped,
    foreign,
    mapped_column,
    relationship,
)

from ai.backend.manager.data.permission.role import (
    RoleData,
    RoleDetailData,
)
from ai.backend.manager.data.permission.status import (
    RoleStatus,
)
from ai.backend.manager.data.permission.types import RoleSource
from ai.backend.manager.models.base import (
    GUID,
    Base,
    StrEnumType,
)

if TYPE_CHECKING:
    from .permission.object_permission import ObjectPermissionRow
    from .user_role import UserRoleRow


def _get_mapped_user_role_rows_join_condition() -> sa.ColumnElement[bool]:
    from .user_role import UserRoleRow

    return RoleRow.id == foreign(UserRoleRow.role_id)


def _get_object_permission_rows_join_condition() -> sa.ColumnElement[bool]:
    from .permission.object_permission import ObjectPermissionRow

    return RoleRow.id == foreign(ObjectPermissionRow.role_id)


class RoleRow(Base):  # type: ignore[misc]
    __tablename__ = "roles"
    __table_args__ = (sa.Index("ix_id_status", "id", "status"),)

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    name: Mapped[str] = mapped_column("name", sa.String(64), nullable=False)
    description: Mapped[str | None] = mapped_column("description", sa.Text, nullable=True)
    source: Mapped[RoleSource] = mapped_column(
        "source",
        StrEnumType(RoleSource, length=16),
        nullable=False,
        default=RoleSource.SYSTEM,
        server_default=str(RoleSource.SYSTEM),
    )
    status: Mapped[RoleStatus] = mapped_column(
        "status",
        StrEnumType(RoleStatus),
        nullable=False,
        default=RoleStatus.ACTIVE,
        server_default=RoleStatus.ACTIVE,
    )
    created_at: Mapped[datetime] = mapped_column(
        "created_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()
    )
    updated_at: Mapped[datetime | None] = mapped_column(
        "updated_at", sa.DateTime(timezone=True), nullable=True
    )
    deleted_at: Mapped[datetime | None] = mapped_column(
        "deleted_at", sa.DateTime(timezone=True), nullable=True
    )

    mapped_user_role_rows: Mapped[list[UserRoleRow]] = relationship(
        "UserRoleRow",
        back_populates="role_row",
        primaryjoin=_get_mapped_user_role_rows_join_condition,
    )
    object_permission_rows: Mapped[list[ObjectPermissionRow]] = relationship(
        "ObjectPermissionRow",
        back_populates="role_row",
        primaryjoin=_get_object_permission_rows_join_condition,
        viewonly=True,
    )

    def to_data(self) -> RoleData:
        return RoleData(
            id=self.id,
            name=self.name,
            source=self.source,
            status=self.status,
            created_at=self.created_at,
            updated_at=self.updated_at or self.created_at,
            deleted_at=self.deleted_at,
            description=self.description,
        )

    def to_detail_data_without_users(self) -> RoleDetailData:
        """Convert to detail data without assigned users."""
        return RoleDetailData(
            id=self.id,
            name=self.name,
            source=self.source,
            status=self.status,
            created_at=self.created_at,
            updated_at=self.updated_at or self.created_at,
            deleted_at=self.deleted_at,
            description=self.description,
            object_permissions=[op_row.to_data() for op_row in self.object_permission_rows],
        )



"""CreatorSpec implementations for object storage domain."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.models.object_storage import ObjectStorageRow
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class ObjectStorageCreatorSpec(CreatorSpec[ObjectStorageRow]):
    """CreatorSpec for object storage creation."""

    name: str
    host: str
    access_key: str
    secret_key: str
    endpoint: str
    region: str

    @override
    def build_row(self) -> ObjectStorageRow:
        return ObjectStorageRow(
            name=self.name,
            host=self.host,
            access_key=self.access_key,
            secret_key=self.secret_key,
            endpoint=self.endpoint,
            region=self.region,
        )



from __future__ import annotations

import logging
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship

from ai.backend.common.data.artifact.types import VerificationStepResult
from ai.backend.common.data.storage.registries.types import ModelData
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.artifact.types import (
    ArtifactRemoteStatus,
    ArtifactRevisionData,
    ArtifactStatus,
)
from ai.backend.manager.models.base import (
    GUID,
    Base,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.artifact import ArtifactRow
    from ai.backend.manager.models.association_artifacts_storages import (
        AssociationArtifactsStorageRow,
    )

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("ArtifactRevisionRow",)


def _get_artifact_join_cond() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.artifact import ArtifactRow

    return foreign(ArtifactRevisionRow.artifact_id) == ArtifactRow.id


def _get_association_artifacts_storages_join_cond() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.association_artifacts_storages import (
        AssociationArtifactsStorageRow,
    )

    return ArtifactRevisionRow.id == foreign(AssociationArtifactsStorageRow.artifact_revision_id)


class ArtifactRevisionRow(Base):  # type: ignore[misc]
    __tablename__ = "artifact_revisions"
    __table_args__ = (
        # constraint
        sa.UniqueConstraint("artifact_id", "version", name="uq_artifact_id_version"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    artifact_id: Mapped[uuid.UUID] = mapped_column(
        "artifact_id",
        GUID,
        nullable=False,
        index=True,
    )
    version: Mapped[str] = mapped_column("version", sa.String, nullable=False)
    readme: Mapped[str | None] = mapped_column("readme", sa.TEXT, nullable=True, default=None)
    size: Mapped[int | None] = mapped_column("size", sa.BigInteger, nullable=True, default=None)
    digest: Mapped[str | None] = mapped_column(
        "digest", sa.String, nullable=True, server_default=None, default=None
    )

    # It's unnatural to include "status" in the revision, but let's put it here for now instead of creating separate table.
    status: Mapped[str] = mapped_column(
        "status", sa.String, index=True, nullable=False, default=ArtifactStatus.SCANNED.value
    )
    remote_status: Mapped[str | None] = mapped_column(
        "remote_status",
        sa.String,
        index=True,
        nullable=True,
        default=None,
        server_default=sa.null(),
    )

    created_at: Mapped[datetime | None] = mapped_column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=True,
        server_default=None,
        index=True,
    )
    updated_at: Mapped[datetime | None] = mapped_column(
        "updated_at",
        sa.DateTime(timezone=True),
        nullable=True,
        server_default=None,
        index=True,
    )
    verification_result: Mapped[dict[str, Any] | None] = mapped_column(
        "verification_result", sa.JSON(none_as_null=True), nullable=True, default=None
    )

    artifact: Mapped[ArtifactRow] = relationship(
        "ArtifactRow",
        back_populates="revision_rows",
        primaryjoin=_get_artifact_join_cond,
        viewonly=True,
    )

    association_artifacts_storages_rows: Mapped[list[AssociationArtifactsStorageRow]] = (
        relationship(
            "AssociationArtifactsStorageRow",
            back_populates="artifact_revision_row",
            primaryjoin=_get_association_artifacts_storages_join_cond,
        )
    )

    def __str__(self) -> str:
        readme_display = self.readme[:15] if self.readme else None
        created_at_str = self.created_at.isoformat() if self.created_at else None
        updated_at_str = self.updated_at.isoformat() if self.updated_at else None
        return (
            f"ArtifactRevisionRow("
            f"id={self.id}, "
            f"artifact_id={self.artifact_id}, "
            f"version={self.version}, "
            f"readme={readme_display}, "
            f"size={self.size}, "
            f"status={self.status}, "
            f"remote_status={self.remote_status}, "
            f"created_at={created_at_str}, "
            f"updated_at={updated_at_str}, "
            f"digest={self.digest}"
            f")"
        )

    def to_dataclass(self) -> ArtifactRevisionData:
        # Convert JSON dict back to Pydantic model if present
        verification_result = None
        if self.verification_result is not None:
            try:
                verification_result = VerificationStepResult.model_validate(
                    self.verification_result
                )
            except Exception as e:
                # If validation fails, keep as None
                verification_result = None
                log.warning(
                    "Failed to validate verification_result for ArtifactRevisionRow id={}: {}",
                    self.id,
                    e,
                )

        return ArtifactRevisionData(
            id=self.id,
            artifact_id=self.artifact_id,
            version=self.version,
            readme=self.readme,
            size=self.size,
            status=ArtifactStatus(self.status),
            remote_status=ArtifactRemoteStatus(self.remote_status) if self.remote_status else None,
            created_at=self.created_at,
            updated_at=self.updated_at,
            digest=self.digest,
            verification_result=verification_result,
        )

    @classmethod
    def from_huggingface_model_data(
        cls,
        artifact_id: uuid.UUID,
        model_data: ModelData,
    ) -> ArtifactRevisionRow:
        return cls(
            artifact_id=artifact_id,
            version=model_data.revision,
            readme=model_data.readme,
            size=model_data.size,
            status=ArtifactStatus.SCANNED.value,
            remote_status=None,
            created_at=model_data.created_at,
            updated_at=model_data.modified_at,
            digest=model_data.sha,
            verification_result=None,
        )



"""
ImageV2 GQL types for Strawberry GraphQL.

This module provides ImageV2 types as part of the Strawberry GraphQL migration (BEP-1010).
See BEP-1038 for detailed specifications.
"""

from .fetcher import fetch_image, fetch_image_alias, fetch_image_aliases, fetch_images
from .resolver import (
    admin_image_aliases,
    admin_images_v2,
    container_registry_images_v2,
    image_alias,
    image_scoped_aliases,
    image_v2,
)
from .types import (
    ContainerRegistryScopeGQL,
    ImageV2AliasConnectionGQL,
    ImageV2AliasEdgeGQL,
    ImageV2AliasFilterGQL,
    ImageV2AliasGQL,
    ImageV2AliasOrderByGQL,
    ImageV2AliasOrderFieldGQL,
    ImageV2ConnectionGQL,
    ImageV2EdgeGQL,
    ImageV2FilterGQL,
    ImageV2GQL,
    ImageV2IdentityInfoGQL,
    ImageV2LabelEntryGQL,
    ImageV2MetadataInfoGQL,
    ImageV2OrderByGQL,
    ImageV2OrderFieldGQL,
    ImageV2PermissionGQL,
    ImageV2PermissionInfoGQL,
    ImageV2RequirementsInfoGQL,
    ImageV2ResourceLimitGQL,
    ImageV2ScopeGQL,
    ImageV2StatusGQL,
    ImageV2TagEntryGQL,
)

__all__ = [
    # Enums
    "ImageV2StatusGQL",
    "ImageV2PermissionGQL",
    "ImageV2OrderFieldGQL",
    "ImageV2AliasOrderFieldGQL",
    # Sub-Info Types
    "ImageV2TagEntryGQL",
    "ImageV2LabelEntryGQL",
    "ImageV2ResourceLimitGQL",
    # Info Types
    "ImageV2IdentityInfoGQL",
    "ImageV2MetadataInfoGQL",
    "ImageV2RequirementsInfoGQL",
    "ImageV2PermissionInfoGQL",
    # Main Types
    "ImageV2GQL",
    "ImageV2EdgeGQL",
    "ImageV2ConnectionGQL",
    "ImageV2AliasGQL",
    "ImageV2AliasEdgeGQL",
    "ImageV2AliasConnectionGQL",
    # Filter, OrderBy, and Scope Types
    "ImageV2FilterGQL",
    "ImageV2OrderByGQL",
    "ImageV2AliasFilterGQL",
    "ImageV2AliasOrderByGQL",
    "ContainerRegistryScopeGQL",
    "ImageV2ScopeGQL",
    # Fetcher functions
    "fetch_images",
    "fetch_image",
    "fetch_image_alias",
    "fetch_image_aliases",
    # Resolver fields
    "admin_images_v2",
    "image_v2",
    "container_registry_images_v2",
    "image_alias",
    "image_scoped_aliases",
    "admin_image_aliases",
]



from __future__ import annotations

import logging
import uuid
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.models.base import GUID, Base

if TYPE_CHECKING:
    from ai.backend.manager.models.artifact_revision import ArtifactRevisionRow
    from ai.backend.manager.models.object_storage import ObjectStorageRow
    from ai.backend.manager.models.vfs_storage import VFSStorageRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__: Sequence[str] = ("AssociationArtifactsStorageRow",)


def _get_association_artifact_join_cond() -> sa.sql.elements.ColumnElement[Any]:
    from ai.backend.manager.models.artifact_revision import ArtifactRevisionRow

    return ArtifactRevisionRow.id == foreign(AssociationArtifactsStorageRow.artifact_revision_id)


def _get_association_object_storage_join_cond() -> sa.sql.elements.ColumnElement[Any]:
    from ai.backend.manager.models.object_storage import ObjectStorageRow

    return ObjectStorageRow.id == foreign(AssociationArtifactsStorageRow.storage_namespace_id)


def _get_association_vfs_storage_join_cond() -> sa.sql.elements.ColumnElement[Any]:
    from ai.backend.manager.models.vfs_storage import VFSStorageRow

    return VFSStorageRow.id == foreign(AssociationArtifactsStorageRow.storage_namespace_id)


class AssociationArtifactsStorageRow(Base):  # type: ignore[misc]
    """
    Association table for linking artifacts to storage namespace.
    """

    __tablename__ = "association_artifacts_storages"
    __table_args__ = (
        # constraint
        sa.UniqueConstraint("artifact_revision_id", name="uq_artifact_revision_id"),
    )

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    artifact_revision_id: Mapped[uuid.UUID] = mapped_column(
        "artifact_revision_id",
        GUID,
        nullable=False,
    )
    storage_namespace_id: Mapped[uuid.UUID] = mapped_column(
        "storage_namespace_id",
        GUID,
        nullable=False,
    )
    storage_type: Mapped[str] = mapped_column("storage_type", sa.String, nullable=False)

    artifact_revision_row: Mapped[ArtifactRevisionRow] = relationship(
        "ArtifactRevisionRow",
        back_populates="association_artifacts_storages_rows",
        primaryjoin=_get_association_artifact_join_cond,
    )

    # only valid when storage_type is "object_storage"
    object_storage_row: Mapped[ObjectStorageRow | None] = relationship(
        "ObjectStorageRow",
        back_populates="association_artifacts_storages_rows",
        primaryjoin=_get_association_object_storage_join_cond,
        overlaps="vfs_storage_row",
    )

    # only valid when storage_type is "vfs"
    vfs_storage_row: Mapped[VFSStorageRow | None] = relationship(
        "VFSStorageRow",
        back_populates="association_artifacts_storages_rows",
        primaryjoin=_get_association_vfs_storage_join_cond,
        overlaps="object_storage_row",
    )



from __future__ import annotations

from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Self

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ai.backend.common.types import (
    DefaultForUnspecified,
    ResourceSlot,
    VFolderHostPermissionMap,
)
from ai.backend.manager.data.resource.types import (
    KeyPairResourcePolicyData,
    ProjectResourcePolicyData,
    UserResourcePolicyData,
)
from ai.backend.manager.models.base import (
    Base,
    EnumType,
    ResourceSlotColumn,
    VFolderHostPermissionColumn,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.group import GroupRow
    from ai.backend.manager.models.keypair import KeyPairRow
    from ai.backend.manager.models.user import UserRow

__all__: Sequence[str] = (
    "DefaultForUnspecified",
    "KeyPairResourcePolicyRow",
    "ProjectResourcePolicyRow",
    "UserResourcePolicyRow",
    "keypair_resource_policies",
    "project_resource_policies",
    "user_resource_policies",
)


class KeyPairResourcePolicyRow(Base):  # type: ignore[misc]
    __tablename__ = "keypair_resource_policies"

    name: Mapped[str] = mapped_column("name", sa.String(length=256), primary_key=True)
    created_at: Mapped[datetime | None] = mapped_column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.func.now()
    )
    default_for_unspecified: Mapped[DefaultForUnspecified] = mapped_column(
        "default_for_unspecified",
        EnumType(DefaultForUnspecified),
        default=DefaultForUnspecified.LIMITED,
        nullable=False,
    )
    total_resource_slots: Mapped[ResourceSlot] = mapped_column(
        "total_resource_slots", ResourceSlotColumn(), nullable=False
    )
    max_session_lifetime: Mapped[int] = mapped_column(
        "max_session_lifetime", sa.Integer(), nullable=False, server_default=sa.text("0")
    )
    max_concurrent_sessions: Mapped[int] = mapped_column(
        "max_concurrent_sessions", sa.Integer(), nullable=False
    )
    max_pending_session_count: Mapped[int | None] = mapped_column(
        "max_pending_session_count", sa.Integer(), nullable=True
    )
    max_pending_session_resource_slots: Mapped[ResourceSlot | None] = mapped_column(
        "max_pending_session_resource_slots", ResourceSlotColumn(), nullable=True
    )
    max_concurrent_sftp_sessions: Mapped[int] = mapped_column(
        "max_concurrent_sftp_sessions", sa.Integer(), nullable=False, server_default=sa.text("1")
    )
    max_containers_per_session: Mapped[int] = mapped_column(
        "max_containers_per_session", sa.Integer(), nullable=False
    )
    idle_timeout: Mapped[int] = mapped_column("idle_timeout", sa.BigInteger(), nullable=False)
    allowed_vfolder_hosts: Mapped[VFolderHostPermissionMap] = mapped_column(
        "allowed_vfolder_hosts",
        VFolderHostPermissionColumn(),
        nullable=False,
        default={},
    )
    # TODO: implement with a many-to-many association table
    # allowed_scaling_groups: Mapped[list[str]] = mapped_column(sa.Array(sa.String), nullable=False)

    keypairs: Mapped[list[KeyPairRow]] = relationship(
        "KeyPairRow", back_populates="resource_policy_row"
    )

    def to_dataclass(
        self,
    ) -> KeyPairResourcePolicyData:
        return KeyPairResourcePolicyData(
            name=self.name,
            created_at=self.created_at,
            default_for_unspecified=self.default_for_unspecified,
            total_resource_slots=self.total_resource_slots,
            max_session_lifetime=self.max_session_lifetime,
            max_concurrent_sessions=self.max_concurrent_sessions,
            max_pending_session_count=self.max_pending_session_count,
            max_pending_session_resource_slots=self.max_pending_session_resource_slots,
            max_concurrent_sftp_sessions=self.max_concurrent_sftp_sessions,
            max_containers_per_session=self.max_containers_per_session,
            idle_timeout=self.idle_timeout,
            allowed_vfolder_hosts=self.allowed_vfolder_hosts,
        )


# NOTE: Deprecated legacy table reference for backward compatibility.
# Use KeyPairResourcePolicyRow class directly for new code.
keypair_resource_policies = KeyPairResourcePolicyRow.__table__


class UserResourcePolicyRow(Base):  # type: ignore[misc]
    __tablename__ = "user_resource_policies"

    name: Mapped[str] = mapped_column("name", sa.String(length=256), primary_key=True)
    created_at: Mapped[datetime | None] = mapped_column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.func.now()
    )
    max_vfolder_count: Mapped[int] = mapped_column(
        "max_vfolder_count", sa.Integer(), nullable=False
    )
    max_quota_scope_size: Mapped[int] = mapped_column(
        "max_quota_scope_size", sa.BigInteger(), nullable=False
    )
    max_session_count_per_model_session: Mapped[int] = mapped_column(
        "max_session_count_per_model_session", sa.Integer(), nullable=False
    )
    max_customized_image_count: Mapped[int] = mapped_column(
        "max_customized_image_count", sa.Integer(), nullable=False, default=3
    )

    users: Mapped[list[UserRow]] = relationship("UserRow", back_populates="resource_policy_row")

    def __init__(
        self,
        name: str,
        max_vfolder_count: int,
        max_quota_scope_size: int,
        max_session_count_per_model_session: int,
        max_customized_image_count: int,
    ) -> None:
        self.name = name
        self.max_vfolder_count = max_vfolder_count
        self.max_quota_scope_size = max_quota_scope_size
        self.max_session_count_per_model_session = max_session_count_per_model_session
        self.max_customized_image_count = max_customized_image_count

    @classmethod
    def from_dataclass(cls, data: UserResourcePolicyData) -> Self:
        return cls(
            name=data.name,
            max_vfolder_count=data.max_vfolder_count,
            max_quota_scope_size=data.max_quota_scope_size,
            max_session_count_per_model_session=data.max_session_count_per_model_session,
            max_customized_image_count=data.max_customized_image_count,
        )

    def to_dataclass(self) -> UserResourcePolicyData:
        return UserResourcePolicyData(
            name=self.name,
            max_vfolder_count=self.max_vfolder_count,
            max_quota_scope_size=self.max_quota_scope_size,
            max_session_count_per_model_session=self.max_session_count_per_model_session,
            max_customized_image_count=self.max_customized_image_count,
        )


# NOTE: Deprecated legacy table reference for backward compatibility.
# Use UserResourcePolicyRow class directly for new code.
user_resource_policies = UserResourcePolicyRow.__table__


class ProjectResourcePolicyRow(Base):  # type: ignore[misc]
    __tablename__ = "project_resource_policies"

    name: Mapped[str] = mapped_column("name", sa.String(length=256), primary_key=True)
    created_at: Mapped[datetime | None] = mapped_column(
        "created_at", sa.DateTime(timezone=True), server_default=sa.func.now()
    )
    max_vfolder_count: Mapped[int] = mapped_column(
        "max_vfolder_count", sa.Integer(), nullable=False
    )
    max_quota_scope_size: Mapped[int] = mapped_column(
        "max_quota_scope_size", sa.BigInteger(), nullable=False
    )
    max_network_count: Mapped[int] = mapped_column(
        "max_network_count", sa.Integer(), nullable=False
    )

    projects: Mapped[list[GroupRow]] = relationship(
        "GroupRow", back_populates="resource_policy_row"
    )

    def __init__(
        self,
        name: str,
        max_vfolder_count: int,
        max_quota_scope_size: int,
        max_network_count: int,
    ) -> None:
        self.name = name
        self.max_vfolder_count = max_vfolder_count
        self.max_quota_scope_size = max_quota_scope_size
        self.max_network_count = max_network_count

    @classmethod
    def from_dataclass(cls, data: ProjectResourcePolicyData) -> Self:
        return cls(
            name=data.name,
            max_vfolder_count=data.max_vfolder_count,
            max_quota_scope_size=data.max_quota_scope_size,
            max_network_count=data.max_network_count,
        )

    def to_dataclass(self) -> ProjectResourcePolicyData:
        return ProjectResourcePolicyData(
            name=self.name,
            max_vfolder_count=self.max_vfolder_count,
            max_quota_scope_size=self.max_quota_scope_size,
            max_network_count=self.max_network_count,
        )


# NOTE: Deprecated legacy table reference for backward compatibility.
# Use ProjectResourcePolicyRow class directly for new code.
project_resource_policies = ProjectResourcePolicyRow.__table__



from .gql_relay import (
    Connection,
)
from .session import ComputeSessionNode


class SessionPendingQueueConnection(Connection):
    class Meta:
        node = ComputeSessionNode
        description = "Added in 25.13.0."



"""
Exceptions for the sokovan scheduler.
"""

from abc import ABC, abstractmethod

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.manager.sokovan.data import SchedulingPredicate


class SchedulingError(BackendAIError, ABC):
    """Base exception for scheduling errors.

    All exceptions used in the scheduler must inherit from this class
    and implement the failed_predicates method.
    """

    error_type = "https://api.backend.ai/probs/scheduling-failed"
    error_title = "Scheduling failed."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.SCHEDULE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )

    @abstractmethod
    def failed_predicates(self) -> list[SchedulingPredicate]:
        """Return list of failed predicates for this error.

        Returns:
            List of SchedulingPredicate objects.
        """
        raise NotImplementedError



from __future__ import annotations

import enum
from collections.abc import Mapping
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any, Self, override

from ai.backend.common.auth import PublicKey
from ai.backend.common.data.agent.types import AgentInfo
from ai.backend.common.types import AgentId, DeviceName, ResourceSlot, SlotName, SlotTypes

if TYPE_CHECKING:
    from ai.backend.manager.models.rbac.permission_defs import AgentPermission


class AgentStatus(enum.Enum):
    ALIVE = 0
    LOST = 1
    RESTARTING = 2
    TERMINATED = 3

    @override
    @classmethod
    def _missing_(cls, value: Any) -> AgentStatus | None:
        if isinstance(value, int):
            for member in cls:
                if member.value == value:
                    return member
        if isinstance(value, str):
            match value.upper():
                case "ALIVE":
                    return cls.ALIVE
                case "LOST":
                    return cls.LOST
                case "RESTARTING":
                    return cls.RESTARTING
                case "TERMINATED":
                    return cls.TERMINATED
        return None

    @classmethod
    def unavailable_statuses(cls) -> tuple[AgentStatus, ...]:
        """
        Return agent statuses that indicate the agent is unavailable for kernel operations.
        These agents cannot handle kernel termination or other operations.
        """
        return (cls.LOST, cls.TERMINATED)


@dataclass
class AgentDataForHeartbeatUpdate:
    status: AgentStatus
    status_changed: datetime | None
    scaling_group: str
    available_slots: ResourceSlot
    addr: str
    public_host: str | None
    version: str
    architecture: str
    compute_plugins: Mapping[str, Any]
    public_key: PublicKey | None
    auto_terminate_abusing_kernel: bool


@dataclass
class AgentData:
    id: AgentId
    status: AgentStatus
    status_changed: datetime | None
    region: str
    scaling_group: str
    schedulable: bool
    available_slots: ResourceSlot
    cached_occupied_slots: ResourceSlot
    actual_occupied_slots: ResourceSlot
    addr: str
    public_host: str | None
    first_contact: datetime | None
    lost_at: datetime | None
    version: str
    architecture: str
    compute_plugins: Mapping[str, Any]
    public_key: PublicKey | None
    auto_terminate_abusing_kernel: bool


@dataclass
class AgentMetadata:
    id: AgentId
    status: AgentStatus
    region: str | None
    scaling_group: str
    architecture: str
    version: str
    auto_terminate_abusing_kernel: bool


@dataclass
class AgentNetworkInfo:
    addr: str
    public_host: str
    public_key: PublicKey | None


@dataclass
class AgentResourceInfo:
    slot_key_and_units: dict[SlotName, SlotTypes]
    available_slots: ResourceSlot
    compute_plugins: Mapping[DeviceName, Mapping[str, str]]


@dataclass
class AgentHeartbeatUpsert:
    metadata: AgentMetadata
    network_info: AgentNetworkInfo
    resource_info: AgentResourceInfo
    lost_at: datetime | None
    heartbeat_received: datetime

    @property
    def insert_fields(self) -> dict[str, Any]:
        return {
            "id": self.metadata.id,
            "status": AgentStatus.ALIVE,
            "region": self.metadata.region,
            "scaling_group": self.metadata.scaling_group,
            "available_slots": self.resource_info.available_slots,
            "addr": self.network_info.addr,
            "public_host": self.network_info.public_host,
            "public_key": self.network_info.public_key,
            "version": self.metadata.version,
            "compute_plugins": self.resource_info.compute_plugins,
            "architecture": self.metadata.architecture,
            "auto_terminate_abusing_kernel": self.metadata.auto_terminate_abusing_kernel,
            "lost_at": None,
            "occupied_slots": ResourceSlot(),
            "first_contact": self.heartbeat_received,
        }

    @property
    def update_fields(self) -> dict[str, Any]:
        return {
            "id": self.metadata.id,
            "status": AgentStatus.ALIVE,
            "region": self.metadata.region,
            "scaling_group": self.metadata.scaling_group,
            "available_slots": self.resource_info.available_slots,
            "addr": self.network_info.addr,
            "public_host": self.network_info.public_host,
            "public_key": self.network_info.public_key,
            "version": self.metadata.version,
            "compute_plugins": self.resource_info.compute_plugins,
            "architecture": self.metadata.architecture,
            "auto_terminate_abusing_kernel": self.metadata.auto_terminate_abusing_kernel,
            "lost_at": None,
        }

    @classmethod
    def from_agent_info(
        cls, agent_id: AgentId, agent_info: AgentInfo, heartbeat_received: datetime
    ) -> Self:
        return cls(
            metadata=AgentMetadata(
                id=agent_id,
                status=AgentStatus.ALIVE,
                region=agent_info.region,
                scaling_group=agent_info.scaling_group,
                architecture=agent_info.architecture,
                auto_terminate_abusing_kernel=agent_info.auto_terminate_abusing_kernel,
                version=agent_info.version,
            ),
            network_info=AgentNetworkInfo(
                addr=agent_info.addr,
                public_host=agent_info.public_host,
                public_key=agent_info.public_key,
            ),
            resource_info=AgentResourceInfo(
                slot_key_and_units=agent_info.slot_key_and_units,
                available_slots=agent_info.available_resource_slots,
                compute_plugins=agent_info.compute_plugins,
            ),
            lost_at=None,
            heartbeat_received=heartbeat_received,
        )


@dataclass
class UpsertResult:
    was_revived: bool
    need_resource_slot_update: bool

    @classmethod
    def from_state_comparison(
        cls, existing_data: AgentDataForHeartbeatUpdate | None, upsert_data: AgentHeartbeatUpsert
    ) -> Self:
        if existing_data is None:
            return cls(
                was_revived=False,
                need_resource_slot_update=True,
            )

        was_revived = existing_data.status in (AgentStatus.LOST, AgentStatus.TERMINATED)
        need_resource_slot_update = (
            existing_data.available_slots != upsert_data.resource_info.available_slots
            or existing_data.scaling_group != upsert_data.metadata.scaling_group
            or existing_data.addr != upsert_data.network_info.addr
            or existing_data.public_host != upsert_data.network_info.public_host
            or existing_data.public_key != upsert_data.network_info.public_key
            or existing_data.version != upsert_data.metadata.version
            or existing_data.compute_plugins != upsert_data.resource_info.compute_plugins
            or existing_data.architecture != upsert_data.metadata.architecture
            or existing_data.auto_terminate_abusing_kernel
            != upsert_data.metadata.auto_terminate_abusing_kernel
        )
        return cls(
            was_revived=was_revived,
            need_resource_slot_update=need_resource_slot_update,
        )


@dataclass
class AgentListResult:
    """Search result with total count for agents."""

    items: list[AgentDetailData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass
class AgentDetailData:
    """
    Agent data with associated permissions.

    This encapsulates an agent's data together with the permissions
    the current user has on that agent.
    """

    agent: AgentData
    permissions: list[AgentPermission]



"""CreatorSpec implementations for deployment domain."""

from .auto_scaling import DeploymentAutoScalingPolicyCreatorSpec
from .deployment import (
    DeploymentCreatorSpec,
    DeploymentExecutionFields,
    DeploymentMetadataFields,
    DeploymentMountFields,
    DeploymentNetworkFields,
    DeploymentReplicaFields,
    DeploymentResourceFields,
    EndpointLifecycleBatchUpdaterSpec,
    ModelRevisionFields,
)
from .policy import DeploymentPolicyCreatorSpec
from .revision import DeploymentRevisionCreatorSpec
from .route import RouteBatchUpdaterSpec, RouteCreatorSpec
from .token import EndpointTokenCreatorSpec

__all__ = [
    # Deployment (endpoint) - CreatorSpec
    "DeploymentCreatorSpec",
    # Deployment (endpoint) - BatchUpdaterSpec
    "EndpointLifecycleBatchUpdaterSpec",
    # Deployment (endpoint) - Field groups
    "DeploymentMetadataFields",
    "DeploymentReplicaFields",
    "DeploymentNetworkFields",
    "DeploymentResourceFields",
    "DeploymentMountFields",
    "DeploymentExecutionFields",
    "ModelRevisionFields",
    # Revision
    "DeploymentRevisionCreatorSpec",
    # Auto-scaling
    "DeploymentAutoScalingPolicyCreatorSpec",
    # Policy
    "DeploymentPolicyCreatorSpec",
    # Route
    "RouteCreatorSpec",
    "RouteBatchUpdaterSpec",
    # Token
    "EndpointTokenCreatorSpec",
]



"""DomainV2 GraphQL API package.

Added in 26.2.0. Provides structured domain management API with typed fields
replacing JSON scalars and organized into logical field groups.
"""

from .resolver import admin_domains_v2, domain_v2, rg_domains_v2
from .types import (
    DomainBasicInfoGQL,
    DomainLifecycleInfoGQL,
    DomainRegistryInfoGQL,
    DomainV2Connection,
    DomainV2Edge,
    DomainV2Filter,
    DomainV2GQL,
    DomainV2OrderBy,
    DomainV2OrderField,
)

__all__ = [
    # Queries
    "domain_v2",
    "admin_domains_v2",
    "rg_domains_v2",
    # Filter and OrderBy
    "DomainV2Filter",
    "DomainV2OrderBy",
    "DomainV2OrderField",
    # Nested types
    "DomainBasicInfoGQL",
    "DomainLifecycleInfoGQL",
    "DomainRegistryInfoGQL",
    # Node types
    "DomainV2GQL",
    "DomainV2Edge",
    "DomainV2Connection",
]



from __future__ import annotations

import uuid
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date, datetime
from enum import StrEnum
from typing import TYPE_CHECKING, Any, Protocol, TypeVar

import graphene
import strawberry
from graphql import StringValueNode
from graphql.language.ast import ValueNode
from graphql_relay.utils import base64, unbase64
from strawberry.relay import Edge, Node
from strawberry.types import get_object_definition, has_object_definition

from ai.backend.manager.data.common.types import IntFilterData, SearchResult, StringFilterData

if TYPE_CHECKING:
    from ai.backend.manager.repositories.base import QueryCondition
    from ai.backend.manager.types import (
        PaginationOptions,
    )


@dataclass(frozen=True)
class StringMatchSpec:
    """Specification for string matching operations in query conditions."""

    value: str
    case_insensitive: bool
    negated: bool


@dataclass(frozen=True)
class UUIDEqualMatchSpec:
    """Specification for UUID equality operations (=, !=)."""

    value: uuid.UUID
    negated: bool


@dataclass(frozen=True)
class UUIDInMatchSpec:
    """Specification for UUID IN operations (IN, NOT IN)."""

    values: list[uuid.UUID]
    negated: bool


@strawberry.scalar
class ByteSize(str):
    """
    Custom scalar type for representing byte sizes in GraphQL.
    """

    @staticmethod
    def parse_value(value: str) -> str:
        return value

    @staticmethod
    def parse_literal(ast: ValueNode) -> str:
        if not isinstance(ast, StringValueNode):
            raise ValueError("ByteSize must be provided as a string literal")
        return ast.value


@strawberry.input
class StringFilter:
    # Basic operations
    contains: str | None = None
    starts_with: str | None = None
    ends_with: str | None = None
    equals: str | None = None

    # NOT operations
    not_contains: str | None = None
    not_starts_with: str | None = None
    not_ends_with: str | None = None
    not_equals: str | None = None

    # Case-insensitive operations
    i_contains: str | None = strawberry.field(name="iContains", default=None)
    i_starts_with: str | None = strawberry.field(name="iStartsWith", default=None)
    i_ends_with: str | None = strawberry.field(name="iEndsWith", default=None)
    i_equals: str | None = strawberry.field(name="iEquals", default=None)

    # Case-insensitive NOT operations
    i_not_contains: str | None = strawberry.field(name="iNotContains", default=None)
    i_not_starts_with: str | None = strawberry.field(name="iNotStartsWith", default=None)
    i_not_ends_with: str | None = strawberry.field(name="iNotEndsWith", default=None)
    i_not_equals: str | None = strawberry.field(name="iNotEquals", default=None)

    def to_dataclass(self) -> StringFilterData:
        return StringFilterData(
            contains=self.contains,
            starts_with=self.starts_with,
            ends_with=self.ends_with,
            equals=self.equals,
            not_equals=self.not_equals,
            i_contains=self.i_contains,
            i_starts_with=self.i_starts_with,
            i_ends_with=self.i_ends_with,
            i_equals=self.i_equals,
            i_not_equals=self.i_not_equals,
        )

    def build_query_condition(
        self,
        contains_factory: Callable[[StringMatchSpec], QueryCondition],
        equals_factory: Callable[[StringMatchSpec], QueryCondition],
        starts_with_factory: Callable[[StringMatchSpec], QueryCondition],
        ends_with_factory: Callable[[StringMatchSpec], QueryCondition],
    ) -> QueryCondition | None:
        """Build a query condition from this filter using the provided factory callables.

        Args:
            contains_factory: Factory for LIKE '%value%' operations
            equals_factory: Factory for exact match (=) operations
            starts_with_factory: Factory for LIKE 'value%' operations
            ends_with_factory: Factory for LIKE '%value' operations

        Returns:
            QueryCondition if any filter field is set, None otherwise
        """
        # equals operations
        if self.equals:
            return equals_factory(
                StringMatchSpec(self.equals, case_insensitive=False, negated=False)
            )
        if self.i_equals:
            return equals_factory(
                StringMatchSpec(self.i_equals, case_insensitive=True, negated=False)
            )
        if self.not_equals:
            return equals_factory(
                StringMatchSpec(self.not_equals, case_insensitive=False, negated=True)
            )
        if self.i_not_equals:
            return equals_factory(
                StringMatchSpec(self.i_not_equals, case_insensitive=True, negated=True)
            )

        # contains operations
        if self.contains:
            return contains_factory(
                StringMatchSpec(self.contains, case_insensitive=False, negated=False)
            )
        if self.i_contains:
            return contains_factory(
                StringMatchSpec(self.i_contains, case_insensitive=True, negated=False)
            )
        if self.not_contains:
            return contains_factory(
                StringMatchSpec(self.not_contains, case_insensitive=False, negated=True)
            )
        if self.i_not_contains:
            return contains_factory(
                StringMatchSpec(self.i_not_contains, case_insensitive=True, negated=True)
            )

        # starts_with operations
        if self.starts_with:
            return starts_with_factory(
                StringMatchSpec(self.starts_with, case_insensitive=False, negated=False)
            )
        if self.i_starts_with:
            return starts_with_factory(
                StringMatchSpec(self.i_starts_with, case_insensitive=True, negated=False)
            )
        if self.not_starts_with:
            return starts_with_factory(
                StringMatchSpec(self.not_starts_with, case_insensitive=False, negated=True)
            )
        if self.i_not_starts_with:
            return starts_with_factory(
                StringMatchSpec(self.i_not_starts_with, case_insensitive=True, negated=True)
            )

        # ends_with operations
        if self.ends_with:
            return ends_with_factory(
                StringMatchSpec(self.ends_with, case_insensitive=False, negated=False)
            )
        if self.i_ends_with:
            return ends_with_factory(
                StringMatchSpec(self.i_ends_with, case_insensitive=True, negated=False)
            )
        if self.not_ends_with:
            return ends_with_factory(
                StringMatchSpec(self.not_ends_with, case_insensitive=False, negated=True)
            )
        if self.i_not_ends_with:
            return ends_with_factory(
                StringMatchSpec(self.i_not_ends_with, case_insensitive=True, negated=True)
            )

        return None


@strawberry.input
class IntFilter:
    equals: int | None = None
    not_equals: int | None = None
    greater_than: int | None = None
    greater_than_or_equal: int | None = None
    less_than: int | None = None
    less_than_or_equal: int | None = None

    def to_dataclass(self) -> IntFilterData:
        return IntFilterData(
            equals=self.equals,
            not_equals=self.not_equals,
            greater_than=self.greater_than,
            greater_than_or_equal=self.greater_than_or_equal,
            less_than=self.less_than,
            less_than_or_equal=self.less_than_or_equal,
        )


@strawberry.input(description="Added in 26.1.0. Filter for UUID fields.")
class UUIDFilter:
    # Basic operations
    equals: uuid.UUID | None = None
    in_: list[uuid.UUID] | None = strawberry.field(name="in", default=None)

    # NOT operations
    not_equals: uuid.UUID | None = None
    not_in: list[uuid.UUID] | None = None

    def build_query_condition(
        self,
        equals_factory: Callable[[UUIDEqualMatchSpec], QueryCondition],
        in_factory: Callable[[UUIDInMatchSpec], QueryCondition],
    ) -> QueryCondition | None:
        """Build a query condition from this filter using the provided factory callables.

        Args:
            equals_factory: Factory function for equality operations (=, !=)
            in_factory: Factory function for IN operations (IN, NOT IN)

        Returns:
            QueryCondition if any filter field is set, None otherwise
        """
        # Equality operations
        if self.equals:
            return equals_factory(
                UUIDEqualMatchSpec(
                    value=self.equals,
                    negated=False,
                )
            )
        if self.not_equals:
            return equals_factory(
                UUIDEqualMatchSpec(
                    value=self.not_equals,
                    negated=True,
                )
            )

        # IN operations
        if self.in_:
            return in_factory(
                UUIDInMatchSpec(
                    values=self.in_,
                    negated=False,
                )
            )
        if self.not_in:
            return in_factory(
                UUIDInMatchSpec(
                    values=self.not_in,
                    negated=True,
                )
            )

        return None


@strawberry.input
class DateTimeFilter:
    """Filter for datetime fields."""

    before: datetime | None = None
    after: datetime | None = None
    equals: datetime | None = None
    not_equals: datetime | None = None

    def build_query_condition(
        self,
        before_factory: Callable[[datetime], QueryCondition],
        after_factory: Callable[[datetime], QueryCondition],
        equals_factory: Callable[[datetime], QueryCondition] | None = None,
    ) -> QueryCondition | None:
        """Build a query condition from this filter using the provided factory callables.

        Args:
            before_factory: Factory function that takes datetime and returns QueryCondition for < comparison
            after_factory: Factory function that takes datetime and returns QueryCondition for > comparison
            equals_factory: Optional factory function for = comparison

        Returns:
            QueryCondition if any filter field is set, None otherwise
        """
        if self.equals and equals_factory:
            return equals_factory(self.equals)
        if self.before:
            return before_factory(self.before)
        if self.after:
            return after_factory(self.after)
        return None


@strawberry.input
class DateFilter:
    """Filter for date fields (not datetime)."""

    before: date | None = None
    after: date | None = None
    equals: date | None = None
    not_equals: date | None = None

    def build_query_condition(
        self,
        before_factory: Callable[[date], QueryCondition],
        after_factory: Callable[[date], QueryCondition],
        equals_factory: Callable[[date], QueryCondition],
        not_equals_factory: Callable[[date], QueryCondition],
    ) -> QueryCondition | None:
        """Build query condition using factory callables.

        Args:
            before_factory: Factory function for < comparison
            after_factory: Factory function for > comparison
            equals_factory: Factory function for = comparison
            not_equals_factory: Factory function for != comparison

        Returns:
            QueryCondition if any filter field is set, None otherwise
        """
        if self.not_equals:
            return not_equals_factory(self.not_equals)
        if self.equals:
            return equals_factory(self.equals)
        if self.before:
            return before_factory(self.before)
        if self.after:
            return after_factory(self.after)
        return None


@strawberry.enum
class OrderDirection(StrEnum):
    ASC = "ASC"
    DESC = "DESC"


@strawberry.enum
class Ordering(StrEnum):
    ASC = "ASC"
    ASC_NULLS_FIRST = "ASC_NULLS_FIRST"
    ASC_NULLS_LAST = "ASC_NULLS_LAST"
    DESC = "DESC"
    DESC_NULLS_FIRST = "DESC_NULLS_FIRST"
    DESC_NULLS_LAST = "DESC_NULLS_LAST"


def to_global_id(
    type_: type[Any], local_id: uuid.UUID | str, is_target_graphene_object: bool = False
) -> str:
    if is_target_graphene_object:
        # For compatibility with existing Graphene-based global IDs
        if not issubclass(type_, graphene.ObjectType):
            raise TypeError(
                "type_ must be a graphene ObjectType when is_target_graphene_object is True."
            )
        typename = type_.__name__
        return base64(f"{typename}:{local_id}")
    if not has_object_definition(type_):
        raise TypeError("type_ must be a Strawberry object type (Node or Edge).")
    typename = get_object_definition(type_, strict=True).name
    return base64(f"{typename}:{local_id}")


def resolve_global_id(global_id: str) -> tuple[str, str]:
    unbased_global_id = unbase64(global_id)
    type_, _, id_ = unbased_global_id.partition(":")
    return type_, id_


CURSOR_VERSION = "v1"


def encode_cursor(row_id: str | uuid.UUID) -> str:
    """Encode row ID to cursor format: base64(cursor:v1:{row_id})"""
    raw = f"cursor:{CURSOR_VERSION}:{row_id}"
    return base64(raw)


def decode_cursor(cursor: str) -> str:
    """Decode cursor and return row_id. Raises InvalidCursor on failure."""
    from ai.backend.manager.errors.api import InvalidCursor

    try:
        raw = unbase64(cursor)
    except Exception as e:
        raise InvalidCursor(f"Invalid cursor encoding: {cursor}") from e

    parts = raw.split(":", 2)
    if len(parts) != 3 or parts[0] != "cursor" or parts[1] != CURSOR_VERSION:
        raise InvalidCursor(f"Invalid cursor format: {cursor}")
    return parts[2]  # row_id


def build_pagination_options(
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> PaginationOptions:
    from ai.backend.manager.types import (
        BackwardPaginationOptions,
        ForwardPaginationOptions,
        OffsetBasedPaginationOptions,
        PaginationOptions,
    )

    """Build pagination options from GraphQL arguments"""
    pagination = PaginationOptions()

    # Handle offset-based pagination
    if offset is not None or limit is not None:
        pagination.offset = OffsetBasedPaginationOptions(offset=offset, limit=limit)
        return pagination

    # Handle cursor-based pagination
    if after is not None or first is not None:
        pagination.forward = ForwardPaginationOptions(after=after, first=first)
    elif before is not None or last is not None:
        pagination.backward = BackwardPaginationOptions(before=before, last=last)

    return pagination


@dataclass
class PageInfo:
    has_next_page: bool
    has_previous_page: bool
    start_cursor: str | None = None
    end_cursor: str | None = None

    def to_strawberry_page_info(self) -> strawberry.relay.PageInfo:
        return strawberry.relay.PageInfo(
            has_next_page=self.has_next_page,
            has_previous_page=self.has_previous_page,
            start_cursor=self.start_cursor,
            end_cursor=self.end_cursor,
        )


class HasCursor(Protocol):
    cursor: str


TEdge = TypeVar("TEdge", bound=HasCursor)


def build_page_info[TEdge: HasCursor](
    edges: list[TEdge], total_count: int, pagination_options: PaginationOptions
) -> PageInfo:
    """Build PageInfo from edges and pagination options"""
    has_next_page = False
    has_previous_page = False

    if pagination_options.offset:
        # Offset-based pagination
        offset = pagination_options.offset.offset or 0

        has_previous_page = offset > 0
        has_next_page = (offset + len(edges)) < total_count

    elif pagination_options.forward:
        # Forward pagination (after/first)
        first = pagination_options.forward.first
        if first is not None:
            # If we got exactly the requested number and there might be more
            has_next_page = len(edges) == first
        else:
            # If no first specified, check if we have all items
            has_next_page = len(edges) < total_count
        has_previous_page = pagination_options.forward.after is not None

    elif pagination_options.backward:
        # Backward pagination (before/last)
        last = pagination_options.backward.last
        if last is not None:
            # If we got exactly the requested number, there might be more before
            has_previous_page = len(edges) == last
        else:
            # If no last specified, assume there could be previous items
            has_previous_page = True
        has_next_page = pagination_options.backward.before is not None

    else:
        # Default case - assume we have all items if no pagination specified
        has_next_page = len(edges) < total_count
        has_previous_page = False

    return PageInfo(
        has_next_page=has_next_page,
        has_previous_page=has_previous_page,
        start_cursor=edges[0].cursor if edges else None,
        end_cursor=edges[-1].cursor if edges else None,
    )


def build_connection[TData, TNode: Node, TConn](
    *,
    result: SearchResult[TData],
    to_node: Callable[[TData], TNode],
    id_getter: Callable[[TNode], str],
    edge_type: type[Edge[TNode]],
    connection_type: Callable[..., TConn],
) -> TConn:
    """Build a relay Connection from a SearchResult."""
    nodes = [to_node(item) for item in result.items]
    edges = [edge_type(node=node, cursor=encode_cursor(id_getter(node))) for node in nodes]
    return connection_type(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=result.total_count,
    )



from __future__ import annotations

import uuid
from datetime import datetime
from typing import (
    TYPE_CHECKING,
)

import sqlalchemy as sa
from sqlalchemy.orm import (
    Mapped,
    foreign,
    mapped_column,
    relationship,
)

from ai.backend.manager.data.permission.role import (
    UserRoleAssignmentData,
    UserRoleAssignmentInput,
)
from ai.backend.manager.models.base import (
    GUID,
    Base,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.user import UserRow

    from .role import RoleRow


def _get_role_row_join_condition() -> sa.ColumnElement[bool]:
    from .role import RoleRow

    return RoleRow.id == foreign(UserRoleRow.role_id)


def _get_user_row_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.user import UserRow

    return UserRow.uuid == foreign(UserRoleRow.user_id)


class UserRoleRow(Base):  # type: ignore[misc]
    __tablename__ = "user_roles"
    __table_args__ = (sa.UniqueConstraint("user_id", "role_id", name="uq_user_id_role_id"),)

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    user_id: Mapped[uuid.UUID] = mapped_column("user_id", GUID, nullable=False)
    role_id: Mapped[uuid.UUID] = mapped_column("role_id", GUID, nullable=False)
    granted_by: Mapped[uuid.UUID | None] = mapped_column(
        "granted_by", GUID, nullable=True
    )  # Null if granted by system
    granted_at: Mapped[datetime] = mapped_column(
        "granted_at", sa.DateTime(timezone=True), nullable=False, server_default=sa.func.now()
    )

    role_row: Mapped[RoleRow | None] = relationship(
        "RoleRow",
        back_populates="mapped_user_role_rows",
        primaryjoin=_get_role_row_join_condition,
    )
    user_row: Mapped[UserRow | None] = relationship(
        "UserRow",
        back_populates="role_assignments",
        primaryjoin=_get_user_row_join_condition,
    )

    def to_data(self) -> UserRoleAssignmentData:
        return UserRoleAssignmentData(
            id=self.id,
            user_id=self.user_id,
            role_id=self.role_id,
            granted_by=self.granted_by,
        )

    @classmethod
    def from_input(cls, input: UserRoleAssignmentInput) -> UserRoleRow:
        return cls(
            user_id=input.user_id,
            role_id=input.role_id,
            granted_by=input.granted_by,
        )
